from cumulusci.tasks.robotframework.debugger.ui import DebuggerCli  # noqa: F401
from cumulusci.tasks.robotframework.debugger.model import (  # noqa: F401
    Breakpoint,
    Suite,
    Testcase,
    Keyword,
)
from cumulusci.tasks.robotframework.debugger.DebugListener import (  # noqa: F401
    DebugListener,
)



"""Utilities for working with MetaDeploy objects"""

from cumulusci.core.config import FlowConfig, TaskConfig
from cumulusci.core.flowrunner import FlowCoordinator
from cumulusci.utils import cd


def get_frozen_steps(project_config, plan_config):
    """Return a list of frozen steps for a MetaDeploy plan config"""
    flow_config = FlowConfig(plan_config)
    flow_config.project_config = project_config
    flow = FlowCoordinator(project_config, flow_config)
    steps = []
    for step in flow.steps:
        if step.skip:
            continue
        with cd(step.project_config.repo_root):
            task = step.task_class(
                step.project_config,
                TaskConfig(step.task_config),
                name=step.task_name,
            )
            steps.extend(task.freeze(step))
    return steps



from cumulusci.core.utils import process_bool_arg
from cumulusci.tasks.metadata.package import PackageXmlGenerator
from cumulusci.tasks.salesforce import UninstallLocalBundles


class UninstallLocalNamespacedBundles(UninstallLocalBundles):

    task_options = {
        "path": {
            "description": "The path to a directory containing the metadata bundles (subdirectories) to uninstall",
            "required": True,
        },
        "managed": {
            "description": "If True, will insert the actual namespace prefix.  Defaults to False or no namespace"
        },
        "namespace": {
            "description": "The namespace to replace the token with if in managed mode. Defaults to project__package__namespace"
        },
        "filename_token": {
            "description": "The path to the parent directory containing the metadata bundles directories",
            "required": True,
        },
        "purge_on_delete": {
            "description": "Sets the purgeOnDelete option for the deployment.  Defaults to True",
            "required": True,
        },
    }

    def _init_options(self, kwargs):
        super(UninstallLocalNamespacedBundles, self)._init_options(kwargs)

        self.options["managed"] = process_bool_arg(self.options.get("managed", False))
        if "namespace" not in self.options:
            self.options["namespace"] = self.project_config.project__package__namespace
        self.options["purge_on_delete"] = process_bool_arg(
            self.options.get("purge_on_delete", True)
        )

    def _get_destructive_changes(self, path=None):
        if not path:
            path = self.options["path"]

        generator = PackageXmlGenerator(
            directory=path,
            api_version=self.project_config.project__package__api_version,
            delete=True,
        )
        namespace = ""
        if self.options["managed"]:
            if self.options["namespace"]:
                namespace = self.options["namespace"] + "__"

        destructive_changes = generator()
        destructive_changes = destructive_changes.replace(
            self.options["filename_token"], namespace
        )

        return destructive_changes



import json
from unittest import mock

import click
import pytest

from cumulusci.cli.runtime import CliRuntime

from .. import plan
from .utils import run_click_command


@pytest.fixture
def runtime():
    runtime = CliRuntime()
    runtime.project_config.config["plans"] = {
        "plan 1": {
            "title": "Test Plan #1",
            "slug": "plan1_slug",
            "tier": "primary",
            "preflight_message": "This is a preflight message",
            "error_message": "This is an error message",
            "steps": {
                1: {
                    "task": "run_tests",
                    "ui_options": {
                        "name": "Run Tests",
                        "is_recommended": False,
                        "is_required": False,
                    },
                    "checks": [
                        {
                            "when": "soon",
                            "action": "error",
                            "message": "Danger Will Robinson!",
                        }
                    ],
                }
            },
            "checks": [
                {
                    "when": "'test package' not in tasks.get_installed_packages()",
                    "action": "error",
                    "message": "Test Package must be installed in your org.",
                }
            ],
        },
        "plan 2": {
            "title": "Test Plan #2",
            "slug": "plan2_slug",
            "tier": "secondary",
        },
        "plan 3": {
            "title": "Test Plan #3",
            "slug": "plan3_slug",
            "tier": "additional",
            "group": "optional",
        },
    }

    yield runtime


class TestPlanList:
    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_list(self, cli_table, runtime):
        """Happy-path smoke test"""

        run_click_command(plan.plan_list, runtime=runtime, print_json=False)

        cli_table.called_once_with(
            data=[
                ["Name", "Title", "Slug", "Tier"],
                ["plan 1", "Test Plan #1", "plan1_slug", "primary"],
                ["plan 2", "Test Plan #2", "plan2_slug", "secondary"],
                ["plan 3", "Test Plan #3", "plan3_slug", "additional"],
            ]
        )

    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_list__no_plans(self, cli_table, runtime):
        """Make sure we gracefully handle projects with no plans"""
        runtime.project_config.config["plans"] = []

        run_click_command(plan.plan_list, runtime=runtime, print_json=False)

        assert cli_table.call_count == 1
        cli_table.assert_called_once_with(data=[["Name", "Title", "Slug", "Tier"]])

    def test_plan_list__json(self, runtime):
        """Verify we get valid json output with --json"""
        stdout = []
        with mock.patch("click.echo", stdout.append):
            run_click_command(plan.plan_list, runtime=runtime, print_json=True)
            assert len(stdout) == 1
            data = json.loads(stdout[0])
            assert data == [
                {
                    "name": "plan 1",
                    "title": "Test Plan #1",
                    "slug": "plan1_slug",
                    "tier": "primary",
                },
                {
                    "name": "plan 2",
                    "title": "Test Plan #2",
                    "slug": "plan2_slug",
                    "tier": "secondary",
                },
                {
                    "name": "plan 3",
                    "title": "Test Plan #3",
                    "slug": "plan3_slug",
                    "tier": "additional",
                },
            ]

    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_list__sorting(self, cli_table):
        runtime = CliRuntime()
        # define some plans in reverse order from how they
        # should appear...
        runtime.project_config.config["plans"] = {
            "Plan 4": dict(tier="additional"),
            "Plan 3": dict(tier="additional"),
            "Plan 2": dict(tier="secondary"),
            "Plan 1": dict(tier="primary"),
        }

        run_click_command(plan.plan_list, runtime=runtime, print_json=False)
        cli_table.assert_called_once_with(
            data=[
                ["Name", "Title", "Slug", "Tier"],
                ["Plan 1", "", "", "primary"],
                ["Plan 2", "", "", "secondary"],
                ["Plan 3", "", "", "additional"],
                ["Plan 4", "", "", "additional"],
            ],
        )


class TestPlanInfo:
    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_info__config(self, cli_table, runtime):
        """Verify we are creating a CliTable for the config"""
        run_click_command(
            plan.plan_info, "plan 1", runtime=runtime, messages_only=False
        )
        cli_table.assert_any_call(
            title="Config",
            data=[
                ["Key", "Value"],
                ["YAML Key", "plan 1"],
                ["Slug", "plan1_slug"],
                ["Tier", "primary"],
                ["Hidden?", False],
            ],
        ),

    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_info__messages(self, cli_table, runtime):
        """Verify we are creating a CliTable for messages"""
        run_click_command(
            plan.plan_info, "plan 1", runtime=runtime, messages_only=False
        )
        cli_table.assert_any_call(
            title="Messages",
            data=[
                ["Type", "Message"],
                ["Title", "Test Plan #1"],
                ["Preflight", "This is a preflight message"],
                ["Post-install", ""],
                ["Error", "This is an error message"],
            ],
        )

    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_info__preflight_checks(self, cli_table, runtime):
        """Verify we are creating a CliTable for the preflight checks"""
        run_click_command(
            plan.plan_info, "plan 1", runtime=runtime, messages_only=False
        )
        cli_table.assert_any_call(
            title="Plan Preflights",
            data=[
                ["Action", "Message", "When"],
                [
                    "error",
                    "Test Package must be installed in your org.",
                    "'test package' not in tasks.get_installed_packages()",
                ],
            ],
        ),

    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_info__step_preflight_checks(self, cli_table, runtime):
        """Verify we are creating a CliTable for the step preflight checks"""
        run_click_command(
            plan.plan_info, "plan 1", runtime=runtime, messages_only=False
        )
        cli_table.assert_any_call(
            title="Step Preflights",
            data=[
                ["Step", "Action", "Message", "When"],
                [1, "error", "Danger Will Robinson!", "soon"],
            ],
        ),

    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_info__steps(self, cli_table, runtime):
        """Verify we are creating a CliTable for the preflight checks"""
        run_click_command(
            plan.plan_info, "plan 1", runtime=runtime, messages_only=False
        )
        cli_table.assert_any_call(
            title="Steps",
            data=[
                ["Step", "Name", "Required", "Recommended"],
                [1, "Run Tests", False, False],
            ],
        )

    @mock.patch("cumulusci.cli.plan.CliTable")
    def test_plan_info__messages_only(self, cli_table, runtime):
        """Verify that --messages results in only messages being output"""
        run_click_command(plan.plan_info, "plan 1", runtime=runtime, messages_only=True)
        cli_table.assert_called_once_with(
            title="Messages",
            data=[
                ["Type", "Message"],
                ["Title", "Test Plan #1"],
                ["Preflight", "This is a preflight message"],
                ["Post-install", ""],
                ["Error", "This is an error message"],
            ],
        )

    def test_plan_info__bogus_plan(self, runtime):
        """Verify a missing play causes a useful message"""
        with pytest.raises(click.UsageError, match=r"Unknown plan 'invalid_plan'."):
            run_click_command(
                plan.plan_info, "invalid_plan", runtime=runtime, messages_only=False
            )



import pytest


class TestConfigBackwardsCompatibility:
    def test_temporary_backwards_compatiblity_hacks(self):
        with pytest.warns(UserWarning):
            from cumulusci.core.config.OrgConfig import OrgConfig

            assert isinstance(OrgConfig, type)

        with pytest.warns(UserWarning):
            from cumulusci.core.config.ScratchOrgConfig import ScratchOrgConfig  # noqa

            assert isinstance(ScratchOrgConfig, type)

        with pytest.warns(UserWarning):
            from cumulusci.core.config.BaseConfig import BaseConfig

            assert isinstance(BaseConfig, type)

        with pytest.warns(UserWarning):
            from cumulusci.core.config.BaseTaskFlowConfig import BaseTaskFlowConfig

            assert isinstance(BaseTaskFlowConfig, type)



from pathlib import Path

import xmltodict

from cumulusci.core.exceptions import CumulusCIException
from cumulusci.tasks.salesforce import UninstallPackaged
from cumulusci.utils import package_xml_from_dict, temporary_dir

DEFAULT_IGNORE_TYPES = ["RecordType", "CustomObjectTranslation"]


class UninstallPackagedIncremental(UninstallPackaged):
    name = "UninstallPackagedIncremental"
    task_options = {
        "path": {
            "description": "The local path to compare to the retrieved packaged metadata from the org.  Defaults to src.",
            "required": True,
        },
        "package": {
            "description": "The package name to uninstall.  All metadata from the package will be retrieved and a custom destructiveChanges.xml package will be constructed and deployed to delete all deleteable metadata from the package.  Defaults to project__package__name.",
            "required": True,
        },
        "purge_on_delete": {
            "description": "Sets the purgeOnDelete option for the deployment.  Defaults to True.",
            "required": True,
        },
        "ignore": {
            "description": "Components to ignore in the org and not try to delete. Mapping of component type to a list of member names."
        },
        "ignore_types": {
            "description": f"List of component types to ignore in the org and not try to delete. Defaults to {DEFAULT_IGNORE_TYPES}."
        },
        "dry_run": {
            "description": "Perform a dry run of the operation without actually deleting any components, and display the components that would be deleted."
        },
    }

    def _init_options(self, kwargs):
        super(UninstallPackagedIncremental, self)._init_options(kwargs)
        if "path" not in self.options:
            self.options["path"] = "src"
        self.options["ignore"] = self.options.get("ignore", {})
        self.options["ignore_types"] = self.options.get(
            "ignore_types", DEFAULT_IGNORE_TYPES
        )

    def _get_destructive_changes(self, path=None):
        path = path or self.options["path"]
        package_path = Path(path).absolute() / "package.xml"
        if not package_path.is_file():
            raise CumulusCIException(
                "Either you are missing your package.xml file or this "
                "is not a Metadata API format project. You will need to "
                "create a package.xml file or convert your source first "
                "before you can run uninstall_packaged_incremental."
            )

        self.logger.info(
            f"Retrieving metadata in package {self.options['package']} from target org"
        )
        packaged = self._retrieve_packaged()
        with temporary_dir() as tempdir:
            packaged.extractall(tempdir)
            destructive_changes = self._package_xml_diff(
                package_path, Path(tempdir) / "package.xml"
            )

        self.logger.info(
            "Deleting metadata in package {} from target org".format(
                self.options["package"]
            )
            if destructive_changes
            else "No metadata found to delete"
        )
        return destructive_changes

    def _package_xml_diff(self, baseline, compare):
        with open(baseline, "rb") as f:
            baseline_xml = xmltodict.parse(f)
        with open(compare, "rb") as f:
            compare_xml = xmltodict.parse(f)

        delete = {}

        ignore = self.options["ignore"]
        baseline_items = {}
        compare_items = {}
        md_types = baseline_xml["Package"].get("types", [])
        md_types = [md_types] if not isinstance(md_types, list) else md_types
        for md_type in md_types:
            baseline_items[md_type["name"]] = []
            if "members" not in md_type:
                continue
            if isinstance(md_type["members"], str):
                baseline_items[md_type["name"]].append(md_type["members"])
            else:
                baseline_items[md_type["name"]].extend(md_type["members"])

        md_types = compare_xml["Package"].get("types", [])
        md_types = [md_types] if not isinstance(md_types, list) else md_types
        for md_type in md_types:
            compare_items[md_type["name"]] = []
            if "members" not in md_type:
                continue
            if isinstance(md_type["members"], str):
                md_type["members"] = [md_type["members"]]
            for item in md_type["members"]:
                if item in ignore.get(md_type["name"], []):
                    continue
                compare_items[md_type["name"]].append(item)

        for md_type, members in compare_items.items():
            if md_type not in baseline_items:
                delete[md_type] = members
                continue

            for member in members:
                if member not in baseline_items[md_type]:
                    if md_type not in delete:
                        delete[md_type] = []
                    delete[md_type].append(member)

        if delete:
            self.logger.info("Deleting metadata:")
            for skip_type in self.options["ignore_types"]:
                delete.pop(skip_type, None)
            for md_type, members in delete.items():
                for member in members:
                    self.logger.info("    {}: {}".format(md_type, member))
            destructive_changes = self._render_xml_from_items_dict(delete)
            return destructive_changes

    def _render_xml_from_items_dict(self, items):
        return package_xml_from_dict(
            items, api_version=self.project_config.project__package__api_version
        )



import os
import shutil

import fs
from github3.exceptions import NotFoundError

from cumulusci.core.exceptions import DependencyResolutionError
from cumulusci.core.github import (
    catch_common_github_auth_errors,
    find_latest_release,
    find_previous_release,
    get_github_api_for_repo,
)
from cumulusci.utils import download_extract_github
from cumulusci.utils.git import split_repo_url
from cumulusci.utils.yaml.cumulusci_yml import GitHubSourceModel, GitHubSourceRelease


class GitHubSource:
    def __init__(self, project_config, spec: GitHubSourceModel):
        self.project_config = project_config
        self.spec = spec
        self.url = spec.github
        if self.url.endswith(".git"):
            self.url = self.url[:-4]

        self.repo_owner, self.repo_name = split_repo_url(self.url)

        try:
            self.gh = get_github_api_for_repo(
                project_config.keychain, self.repo_owner, self.repo_name
            )
            self.repo = self._get_repository(self.repo_owner, self.repo_name)
        except NotFoundError:
            raise DependencyResolutionError(
                f"We are unable to find the repository at {self.url}. Please make sure the URL is correct, that your GitHub user has read access to the repository, and that your GitHub personal access token includes the “repo” scope."
            )
        self.resolve()

    def __repr__(self):
        return f"<GitHubSource {str(self)}>"

    def __str__(self):
        s = f"GitHub: {self.repo_owner}/{self.repo_name}"
        if self.description:
            s += f" @ {self.description}"
        if self.commit != self.description:
            s += f" ({self.commit})"
        return s

    def __hash__(self):
        return hash((self.url, self.commit))

    @catch_common_github_auth_errors
    def _get_repository(self, repo_owner: str, repo_name: str):
        repo = self.gh.repository(repo_owner, repo_name)
        return repo

    def resolve(self):
        """Resolve a GitHub source into a specific commit.

        The spec must include:
        - github: the URL of the GitHub repository

        It's recommended that the source include:
        - resolution_strategy: [production | preproduction | <strategy-name>]

        The spec may instead be specific about the desired ref or release:
        - commit: a commit hash
        - ref: a Git ref
        - branch: a Git branch
        - tag: a Git tag
        - release: "latest" | "previous" | "latest_beta"

        If none of these are specified, CumulusCI will use the resolution strategy "production"
        to locate the appropriate release or ref.
        """
        ref = None
        # These branches preserve some existing behavior: when a source is set to
        # a specific tag or release, there's no fallback, as there would be
        # if we were subsumed within the dependency resolution machinery.

        # If the user was _not_ specific, we will use the full resolution stack,
        # including fallback behaviors.
        if self.spec.commit:
            self.commit = self.description = self.spec.commit
            return
        elif self.spec.ref:
            ref = self.spec.ref
        elif self.spec.tag:
            ref = "tags/" + self.spec.tag
        elif self.spec.branch:
            ref = "heads/" + self.spec.branch
        elif self.spec.release:
            release = None
            if self.spec.release is GitHubSourceRelease.LATEST:
                release = find_latest_release(self.repo, include_beta=False)
            elif self.spec.release is GitHubSourceRelease.LATEST_BETA:
                release = find_latest_release(self.repo, include_beta=True)
            elif self.spec.release is GitHubSourceRelease.PREVIOUS:
                release = find_previous_release(self.repo)
            if release is None:
                raise DependencyResolutionError(
                    f"Could not find release {self.spec.release}."
                )
            ref = "tags/" + release.tag_name
        else:
            # Avoid circular import issues
            from cumulusci.core.dependencies.dependencies import GitHubDynamicDependency
            from cumulusci.core.dependencies.resolvers import (
                get_resolver_stack,
                resolve_dependency,
            )

            # Use resolution strategies to find the right commit.
            dep = GitHubDynamicDependency(github=self.spec.github)
            resolve_dependency(
                dep,
                self.project_config,
                get_resolver_stack(self.project_config, self.spec.resolution_strategy),
            )
            self.commit = self.description = dep.ref
            return

        self.description = ref[6:] if ref.startswith("heads/") else ref
        self.commit = self.repo.ref(ref).object.sha

    def fetch(self):
        """Fetch the archive of the specified commit and construct its project config."""
        with self.project_config.open_cache(
            fs.path.join("projects", self.repo_name, self.commit)
        ) as path:
            zf = download_extract_github(
                self.gh, self.repo_owner, self.repo_name, ref=self.commit
            )
            try:
                zf.extractall(path)
            except Exception:
                # make sure we don't leave an incomplete cache
                shutil.rmtree(path)
                raise

        project_config = self.project_config.construct_subproject_config(
            repo_info={
                "root": os.path.realpath(path),
                "owner": self.repo_owner,
                "name": self.repo_name,
                "url": self.url,
                "commit": self.commit,
            }
        )
        return project_config

    @property
    def frozenspec(self):
        """Return a spec to reconstruct this source at the current commit"""
        return {
            "github": self.url,
            "commit": self.commit,
            "description": self.description,
        }



import typing as T
from pathlib import Path

from snowfakery.api import COUNT_REPS

from .snowfakery_run_until import RunUntilBase
from .snowfakery_working_directory import SnowfakeryWorkingDirectory


class SubtaskConfigurator:
    def __init__(
        self,
        recipe: Path,
        run_until: RunUntilBase,
        ignore_row_errors: bool,
        bulk_mode: T.Literal["Serial", "Parallel"],
    ):
        self.recipe = recipe
        self.run_until = run_until
        self.ignore_row_errors = ignore_row_errors
        self.bulk_mode = bulk_mode

    # todo: move generate_and_load_initial_portion here

    def data_generator_opts(self, working_dir, recipe_options=None, *args, **kwargs):
        """Generate task options for a data generator"""
        wd = SnowfakeryWorkingDirectory(working_dir)
        name = Path(working_dir).name
        parts = name.rsplit("_", 1)
        batch_size = int(parts[-1])

        # The pid is the tool that ensures that every uniqueid is
        # unique even across different processes.
        #
        # big_ids ensures that the pid is incorporated into the ids.
        plugin_options = {
            "pid": str(wd.index),
            "big_ids": "True",
        }

        return {
            "generator_yaml": str(self.recipe),
            "database_url": wd.database_url,
            "num_records": batch_size,
            "reset_oids": False,
            "continuation_file": wd.continuation_file,
            "num_records_tablename": self.run_until.sobject_name or COUNT_REPS,
            "vars": recipe_options,
            "plugin_options": plugin_options,
        }

    def data_loader_opts(self, working_dir: Path):
        wd = SnowfakeryWorkingDirectory(working_dir)

        options = {
            "mapping": wd.mapping_file,
            "reset_oids": False,
            "database_url": wd.database_url,
            "set_recently_viewed": False,
            "ignore_row_errors": self.ignore_row_errors,
            "bulk_mode": self.bulk_mode
            # don't need to pass loading_rules because they are merged into mapping
        }
        return options



import enum
import re
from typing import Optional, Union

from pydantic import BaseModel

VERSION_RE = re.compile(
    r"^(?P<MajorVersion>\d+)"
    r".(?P<MinorVersion>\d+)"
    r"(\.(?P<PatchVersion>\d+))?"
    r"(\.(?P<BuildNumber>\d+))?"
    r"(( \([bB]eta (?P<BetaNumber>\d+)\))?$|(-Beta_(?P<BetaNumberTag>\d+))?$)"
)


class VersionTypeEnum(str, enum.Enum):
    major = "major"
    minor = "minor"
    patch = "patch"
    build = "build"


class PackageType(str, enum.Enum):
    FIRST_GEN = "1GP"
    SECOND_GEN = "2GP"


class PackageVersionNumber(BaseModel):
    """A Salesforce package version parsed into components,
    that knows how to format itself for user presentation and for tag names."""

    MajorVersion: int = 0
    MinorVersion: int = 0
    PatchVersion: int = 0
    BuildNumber: Union[int, str] = 0
    IsReleased: bool = False

    package_type: PackageType = PackageType.SECOND_GEN

    def format_tag(self, prefix: str) -> str:
        """Format version number as a tag name."""
        return (
            f"{prefix}{self.format()}".replace(" (", "-")
            .replace(")", "")
            .replace(" ", "_")
        )

    def __str__(self):
        return self.format()

    def format(self):
        if self.package_type is PackageType.FIRST_GEN:
            return self.format_1gp()
        else:
            return self.format_2gp()

    def format_2gp(self) -> str:
        """Format version number as a string for 2GP packages"""
        return f"{self.MajorVersion}.{self.MinorVersion}.{self.PatchVersion}.{self.BuildNumber}"

    def format_1gp(self) -> str:
        """Format version number as a string for 1GP packages"""
        patch = f".{self.PatchVersion}" if self.PatchVersion else ""
        beta = f" (Beta {self.BuildNumber})" if self.BuildNumber else ""

        return f"{self.MajorVersion}.{self.MinorVersion}{patch}{beta}"

    @classmethod
    def parse_tag(
        cls,
        s: str,
        prefix_beta: str,
        prefix_prod: str,
        package_type: Optional[PackageType] = None,
    ) -> "PackageVersionNumber":
        if s.startswith(prefix_beta):
            version = s[len(prefix_beta) :]
        elif s.startswith(prefix_prod):
            version = s[len(prefix_prod) :]
        else:
            version = s
        return cls.parse(
            version,
            is_released=s.startswith(prefix_prod),
            package_type=package_type,
        )

    @classmethod
    def parse(
        cls,
        s: str,
        is_released: Optional[bool] = None,
        package_type: Optional[PackageType] = None,
    ) -> "PackageVersionNumber":
        """Parse a version number from a string"""
        match = VERSION_RE.match(s)
        if not match:
            raise ValueError(f"Could not parse version number: {s}")

        first_gen_beta = bool(match.group("BetaNumber") or match.group("BetaNumberTag"))
        build_number = int(
            match.group("BuildNumber")
            or match.group("BetaNumber")
            or match.group("BetaNumberTag")
            or 0
        )
        if package_type is None:
            # .0 is rare, but legal, for a 2GP
            if first_gen_beta or not match.group("BuildNumber"):
                package_type = PackageType.FIRST_GEN
            else:
                package_type = PackageType.SECOND_GEN

        if is_released is None:
            is_released = not first_gen_beta

        return PackageVersionNumber(
            MajorVersion=int(match.group("MajorVersion")),
            MinorVersion=int(match.group("MinorVersion")),
            PatchVersion=int(match.group("PatchVersion") or 0),
            BuildNumber=build_number,
            IsReleased=is_released,
            package_type=package_type,
        )

    def increment(self, version_type: VersionTypeEnum = VersionTypeEnum.build):
        """Construct a new PackageVersionNumber by incrementing the specified component."""
        if self.package_type is not PackageType.SECOND_GEN:
            raise ValueError("Cannot increment the version number of a 1GP package")

        parts = self.dict()
        parts["BuildNumber"] = "NEXT"
        parts["IsReleased"] = False

        if version_type == VersionTypeEnum.major:
            parts["MajorVersion"] += 1
            parts["MinorVersion"] = 0
            parts["PatchVersion"] = 0
        if version_type == VersionTypeEnum.minor:
            parts["MinorVersion"] += 1
            parts["PatchVersion"] = 0
        elif version_type == VersionTypeEnum.patch:
            parts["PatchVersion"] += 1

        return PackageVersionNumber(**parts)


PackageVersionNumber.update_forward_refs()



import copy

from cumulusci.robotframework import locators_52

lex_locators = copy.deepcopy(locators_52.lex_locators)
lex_locators["object_list"] = {
    # Note: this matches the <td> with the checkbutton, not the inner checkbutton
    # This is because clicking the actual checkbutton will throw an error that
    # another element will receive the click.
    "checkbutton": '//tbody/tr[.//*[text()="{}"]]//td[.//input[@type="checkbox"]]',
    "status_info": "//force-list-view-manager-status-info",
}



# IMPORT ORDER MATTERS!

# constants used by MetaCI
FAILED_TO_CREATE_SCRATCH_ORG = "Failed to create scratch org"

from cumulusci.core.config.base_config import BaseConfig

# inherit from BaseConfig


class ConnectedAppOAuthConfig(BaseConfig):
    """Salesforce Connected App OAuth configuration"""

    pass


class FlowConfig(BaseConfig):
    """A flow with its configuration merged"""

    pass


from cumulusci.core.config.org_config import OrgConfig


class ServiceConfig(BaseConfig):
    def __init__(self, config, name=None, keychain=None):
        """Services may need access to a keychain and the alias of their service."""
        super().__init__(config, keychain)


class TaskConfig(BaseConfig):
    """A task with its configuration merged"""

    pass


from cumulusci.core.config.base_task_flow_config import BaseTaskFlowConfig

# inherit from BaseTaskFlowConfig
from cumulusci.core.config.project_config import BaseProjectConfig

# inherit from SfdxOrgConfig
from cumulusci.core.config.scratch_org_config import ScratchOrgConfig

# inherit from OrgConfig
from cumulusci.core.config.sfdx_org_config import SfdxOrgConfig

# inherit from BaseProjectConfig
from cumulusci.core.config.universal_config import UniversalConfig

__all__ = (
    "FAILED_TO_CREATE_SCRATCH_ORG",
    "BaseConfig",
    "ConnectedAppOAuthConfig",
    "FlowConfig",
    "OrgConfig",
    "ServiceConfig",
    "TaskConfig",
    "BaseTaskFlowConfig",
    "BaseProjectConfig",
    "SfdxOrgConfig",
    "ScratchOrgConfig",
    "UniversalConfig",
)



import json

import requests

from .base import BaseMarketingCloudTask


class GetUserInfoTask(BaseMarketingCloudTask):
    """Retrieves user info of pertaining to the currently logged in user.
    Sanitizes and returns the payload in self.return_values."""

    def _run_task(self):
        try:
            payload = self.mc_config.get_user_info()
        except requests.exceptions.HTTPError as e:
            self.logger.error(
                f"Exception occurred fetching user info: {e.response.text}"
            )
            raise

        payload = self._sanitize_payload(payload)
        self.logger.info("Successfully fetched user info.")
        self.logger.info(json.dumps(payload, indent=4))

        self.return_values = payload

    def _sanitize_payload(self, payload: dict) -> dict:
        """Removes any sensitive or non-pertinent information from the payload.
        This currently includes the following top-level portions of the response:
        (1) rest
        (2) application
        (3) permissions

        See the following for a comprehensive response example:
        https://developer.salesforce.com/docs/marketing/marketing-cloud/guide/getUserInfo.html
        """
        del payload["rest"]
        del payload["application"]
        del payload["permissions"]
        return payload



from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class GetPermissionSetAssignments(BaseSalesforceApiTask):
    def _run_task(self):
        query = f"SELECT PermissionSet.Name FROM PermissionSetAssignment WHERE AssigneeId = '{self.org_config.user_id}'"
        self.return_values = [
            result["PermissionSet"]["Name"]
            for result in self.sf.query_all(query)["records"]
        ]
        permsets = "\n".join(self.return_values)
        self.logger.info(f"Found permission sets assigned:\n{permsets}")



import shutil
import time
import typing as T
from collections import defaultdict
from contextlib import contextmanager
from datetime import timedelta
from math import ceil
from pathlib import Path
from queue import Empty
from tempfile import TemporaryDirectory, mkdtemp

import psutil
from snowfakery.api import COUNT_REPS, infer_load_file_path
from snowfakery.cci_mapping_files.declaration_parser import (
    ChannelDeclaration,
    SObjectRuleDeclarationFile,
)

import cumulusci.core.exceptions as exc
from cumulusci.core.config import OrgConfig, TaskConfig
from cumulusci.core.debug import get_debug_mode
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.utils import (
    format_duration,
    process_bool_arg,
    process_list_arg,
    process_list_of_pairs_dict_arg,
)
from cumulusci.tasks.bulkdata.generate_and_load_data_from_yaml import (
    GenerateAndLoadDataFromYaml,
)
from cumulusci.tasks.salesforce import BaseSalesforceApiTask

from .snowfakery_utils.queue_manager import (
    SnowfakeryChannelManager,
    data_loader_new_directory_name,
)
from .snowfakery_utils.snowfakery_run_until import PortionGenerator, determine_run_until
from .snowfakery_utils.snowfakery_working_directory import SnowfakeryWorkingDirectory
from .snowfakery_utils.subtask_configurator import SubtaskConfigurator

# A portion serves the same process in this system as a "batch" in
# other systems. The term "batch" is not used to avoid confusion with
# Salesforce Bulk API 1.0 batches. For example, a portion of 250_000
# Account records would be broken into roughly 25 Salesforce upload
# batches.

# The system starts at the MIN_PORTION_SIZE and grows towards the
# MAX_PORTION_SIZE. This is to prevent the org wasting time waiting
# for the first portions.
MIN_PORTION_SIZE = 2_000
MAX_PORTION_SIZE = 250_000
ERROR_THRESHOLD = (
    0  # TODO v2.1: Allow this to be a percentage of recent records instead
)

# time between "ticks" where the task re-evaluates its progress
# relatively arbitrary trade-off between busy-waiting and adding latency.
WAIT_TIME = 3


class Snowfakery(BaseSalesforceApiTask):

    task_docs = """
    Do a data load with Snowfakery.

    All options are optional.

    The most commonly supplied options are `recipe` and one of the three
    `run_until_...` options.
    """

    task_options = {
        "recipe": {
            "required": True,
            "description": "Path to a Snowfakery recipe file determining what data to generate and load.",
        },
        "run_until_records_in_org": {
            "description": """<sobject>:<count>

      Run the recipe repeatedly until the count of <sobject>
      in the org matches the given <count>.

      For example, `--run_until_records_in_org Account:50_000` means:

      Count the Account records in the org. Let’s say the number
      is 20,000. Thus, we must run the recipe over and
      over again until we generate 30,000 new Account records.
      If the recipe also generates e.g.Contacts, Opportunities or whatever
      else, it generates the appropriate number of them to match.

      Underscores are allowed but optional in big numbers: 2000000
      is the same as 2_000_000.
        """
        },
        "run_until_records_loaded": {
            "description": """<sobject>:<count>

      Run the recipe repeatedly until the number of records of
      <sobject> uploaded in this task execution matches <count>.

      For example, `--run_until_records_loaded Account:50_000` means:

      Run the recipe over and over again
      until we generate 50_000 new Account records. If the recipe
      also generates e.g. Contacts, Opportunities or whatever else, it
      generates the appropriate number of them to match.
        """
        },
        "run_until_recipe_repeated": {
            "description": """Run the recipe <count> times,
            no matter what data is already in the org.

            For example, `--run_until_recipe_repeated 50_000` means
            run the recipe 50_000 times."""
        },
        "working_directory": {"description": "Path for temporary / working files"},
        "loading_rules": {
            "description": "Path to .load.yml file containing rules to use to "
            "load the file. Defaults to `<recipename>.load.yml`. "
            "Multiple files can be comma separated."
        },
        "recipe_options": {
            "required": False,
            "description": """Pass values to override options in the format VAR1:foo,VAR2:bar

             Example: --recipe_options weight:10,color:purple""",
        },
        "bulk_mode": {
            "description": "Set to Serial to serialize everything: data generation, data loading, data ingestion through bulk API. Parallel is the default."
        },
        "drop_missing_schema": {
            "description": "Set to True to skip any missing objects or fields instead of stopping with an error."
        },
        "num_processes": {
            "description": "Number of data generating processes. Defaults to matching the number of CPUs."
        },
        "ignore_row_errors": {
            "description": "Boolean: should we continue loading even after running into row errors? "
            "Defaults to False."
        },
    }

    def _validate_options(self):
        "Validate options before executing the task or before freezing it"
        super()._validate_options()
        # Do not store recipe due to MetaDeploy options freezing
        recipe = self.options.get("recipe")
        recipe = Path(recipe)
        if not recipe.exists():
            raise exc.TaskOptionsError(f"Cannot find recipe `{recipe}`")

        self.num_generator_workers = self.options.get("num_processes", None)
        if self.num_generator_workers is not None:
            self.num_generator_workers = int(self.num_generator_workers)
        self.ignore_row_errors = process_bool_arg(
            self.options.get("ignore_row_errors", False)
        )
        loading_rules = process_list_arg(self.options.get("loading_rules")) or []
        self.loading_rules = [Path(path) for path in loading_rules if path]
        self.recipe_options = process_list_of_pairs_dict_arg(
            self.options.get("recipe_options") or {}
        )
        self.bulk_mode = self.options.get("bulk_mode", "Parallel").title()
        if self.bulk_mode and self.bulk_mode not in ["Serial", "Parallel"]:
            raise TaskOptionsError("bulk_mode must be either Serial or Parallel")

    def _init_channel_configs(self, recipe):
        """The channels describe the 'shape' of the communication

        The normal case is a single, parallelized, bulk channel,
        multi-threaded on client and server, using a single user
        account.

        Using .load.yml you can add more channels, utilizing
        more user accounts which can speed up throughput in
        a few cases.

        This method reads files and options to determine
        what channels should be created later.
        """
        channel_decls = read_channel_declarations(recipe, self.loading_rules)

        if channel_decls:
            self.channel_configs = channel_configs_from_decls(
                channel_decls, self.project_config.keychain
            )
        elif self.bulk_mode == "Serial":
            self.channel_configs = [
                standard_channel_config(
                    self.org_config,
                    self.recipe_options,
                    1,
                    1,
                )
            ]
        else:
            self.channel_configs = [
                standard_channel_config(
                    self.org_config,
                    self.recipe_options,
                    self.num_generator_workers,
                    None,
                )
            ]

    def setup(self):
        """Setup for loading."""
        self.debug_mode = get_debug_mode()
        if not self.num_generator_workers:
            # logical CPUs do not really improve performance of CPU-bound
            # code, so we ignore them.
            self.num_generator_workers = psutil.cpu_count(logical=False)
            if self.debug_mode:
                self.logger.info(f"Using {self.num_generator_workers} workers")

        self.run_until = determine_run_until(self.options, self.sf)
        self.start_time = time.time()
        self.recipe = Path(self.options.get("recipe"))
        self.sobject_counts = defaultdict(RunningTotals)
        self._init_channel_configs(self.recipe)

    ## Todo: Consider when this process runs longer than 2 Hours,
    # what will happen to my sf connection?
    def _run_task(self):
        self.setup()

        portions = PortionGenerator(
            self.run_until.gap,
            MIN_PORTION_SIZE,
            MAX_PORTION_SIZE,
        )

        working_directory = self.options.get("working_directory")
        with self.workingdir_or_tempdir(working_directory) as working_directory:
            self._setup_channels_and_queues(working_directory)
            self.logger.info(f"Working directory is {working_directory}")

            if self.run_until.nothing_to_do:
                self.logger.info(
                    f"Dataload is finished before it started! {self.run_until.nothing_to_do_because}"
                )
                return

            template_path, relevant_sobjects = self._generate_and_load_initial_batch(
                working_directory
            )

            # disable OrgReordCounts for now until it's reliability can be better
            # tested and documented.

            # Retrieve OrgRecordCounts code from
            # https://github.com/SFDO-Tooling/CumulusCI/commit/7d703c44b94e8b21f165e5538c2249a65da0a9eb#diff-54676811961455410c30d9c9405a8f3b9d12a6222a58db9d55580a2da3cfb870R147

            self._loop(
                template_path,
                working_directory,
                None,
                portions,
            )
            self.finish()

    def _setup_channels_and_queues(self, working_directory):
        """Set up all of the channels and queues.

        In particular their directories and the in-memory
        runtime datastructures.

        Each channel can hold multiple queues.
        """
        subtask_configurator = SubtaskConfigurator(
            self.recipe, self.run_until, self.ignore_row_errors, self.bulk_mode
        )
        self.queue_manager = SnowfakeryChannelManager(
            project_config=self.project_config,
            logger=self.logger,
            subtask_configurator=subtask_configurator,
        )
        if len(self.channel_configs) == 1:
            channel = self.channel_configs[0]
            self.queue_manager.add_channel(
                org_config=channel.org_config,
                num_generator_workers=channel.declaration.num_generators,
                num_loader_workers=channel.declaration.num_loaders,
                working_directory=working_directory,
                recipe_options=channel.declaration.recipe_options,
            )
        else:
            self.configure_multiple_channels(working_directory)

    def configure_multiple_channels(self, working_directory):
        """If there is more than one channel (=user account),
        pre-allocate work among them.
        """
        allocated_generator_workers = sum(
            (channel.declaration.num_generators or 0)
            for channel in self.channel_configs
        )
        channels_without_workers = len(
            [
                channel.declaration.num_generators
                for channel in self.channel_configs
                if not channel.declaration.num_generators
            ]
        )
        remaining_generator_workers = (
            self.num_generator_workers - allocated_generator_workers
        )
        num_generators_per_channel = ceil(
            remaining_generator_workers / channels_without_workers
        )
        for idx, channel in enumerate(self.channel_configs):
            if self.debug_mode:
                self.logger.info("Initializing %s", channel)
            channel_wd = working_directory / f"channel_{idx}"
            channel_wd.mkdir()
            recipe_options = channel.merge_recipe_options(self.recipe_options)
            generator_workers = (
                channel.declaration.num_generators or num_generators_per_channel
            )

            self.queue_manager.add_channel(
                org_config=channel.org_config,
                num_generator_workers=generator_workers,
                num_loader_workers=channel.declaration.num_loaders,
                working_directory=channel_wd,
                recipe_options=recipe_options,
            )

    def _loop(
        self,
        template_path,
        tempdir,
        org_record_counts_thread,
        portions: PortionGenerator,
    ):
        """The inner loop that controls when data is generated and when we are done."""
        upload_status = self.get_upload_status(
            portions.next_batch_size,
        )

        while not portions.done(upload_status.total_sets_working_on_or_uploaded):
            if self.debug_mode:
                self.logger.info(f"Working Directory: {tempdir}")

            self.queue_manager.tick(
                upload_status,
                template_path,
                tempdir,
                portions,
                self.get_upload_status,
            )
            self.update_running_totals()
            self.print_running_totals()

            time.sleep(WAIT_TIME)

            upload_status = self._report_status(
                portions.batch_size,
                org_record_counts_thread,
                template_path,
            )

        return upload_status

    def _report_status(
        self,
        batch_size,
        org_record_counts_thread,
        template_path,
    ):
        """Let the user know what is going on."""
        self.logger.info(
            "\n********** PROGRESS *********",
        )

        upload_status = self.get_upload_status(
            batch_size or 0,
        )

        self.logger.info(upload_status._display(detailed=self.debug_mode))

        if upload_status.sets_failed:
            # TODO: this is not sufficiently tested.
            #       commenting it out doesn't break tests
            self.log_failures()

        if upload_status.sets_failed > ERROR_THRESHOLD:
            raise exc.BulkDataException(
                f"Errors exceeded threshold: {upload_status.sets_failed} vs {ERROR_THRESHOLD}"
            )

        # TODO: Retrieve OrgRecordCounts code from
        # https://github.com/SFDO-Tooling/CumulusCI/commit/7d703c44b94e8b21f165e5538c2249a65da0a9eb#diff-54676811961455410c30d9c9405a8f3b9d12a6222a58db9d55580a2da3cfb870R147

        return upload_status

    def update_running_totals(self) -> None:
        """Read and collate result reports from sub-processes/sub-threads

        This is a realtime reporting channel which could, in theory, be updated
        before sub-tasks finish. Currently no sub-tasks are coded to do that.

        The logical next step is to allow LoadData to monitor steps one by
        one or even batches one by one.

        Note that until we implement that, we are paying the complexity
        cost of a real-time channel but not getting the benefits of it.
        """
        while True:
            try:
                results = self.queue_manager.get_results_report()
            except Empty:
                break
            if "results" in results and "step_results" in results["results"]:
                self.update_running_totals_from_load_step_results(results["results"])
            elif "error" in results:
                self.logger.warning(f"Error in load: {results}")
            else:  # pragma: no cover
                self.logger.warning(f"Unexpected message from subtask: {results}")

    def update_running_totals_from_load_step_results(self, results: dict) -> None:
        """'Parse' the results from a load step, to keep track of row errors."""
        for result in results["step_results"].values():
            sobject_name = result["sobject"]
            totals = self.sobject_counts[sobject_name]
            totals.errors += result["total_row_errors"]
            totals.successes += result["records_processed"] - result["total_row_errors"]

    def print_running_totals(self):
        for name, result in self.sobject_counts.items():
            self.logger.info(
                f"       {name}: {result.successes:,} successes, {result.errors:,} errors"
            )

    def finish(self):
        """Wait for jobs to finish"""
        old_message = None
        cooldown = 5
        while not self.queue_manager.check_finished():
            status = self.get_upload_status(0)
            datagen_workers = f"{status.sets_being_generated} data generators, "
            msg = f"Waiting for {datagen_workers}{status.sets_being_loaded} uploads to finish"
            if old_message != msg or cooldown < 1:
                old_message = msg
                self.logger.info(msg)
                self.update_running_totals()
                self.print_running_totals()
                cooldown = 5
            else:
                cooldown -= 1
            time.sleep(WAIT_TIME)

        self.log_failures()

        self.logger.info("")
        self.logger.info(" == Results == ")
        self.update_running_totals()
        self.print_running_totals()
        elapsed = format_duration(timedelta(seconds=time.time() - self.start_time))

        if self.run_until.sobject_name:
            result_msg = f"{self.sobject_counts[self.run_until.sobject_name].successes} {self.run_until.sobject_name} records and associated records"
        else:
            result_msg = f"{self.run_until.target:,} iterations"

        self.logger.info(f"☃ Snowfakery created {result_msg} in {elapsed}.")

    def log_failures(self):
        """Log failures from sub-processes to main process"""
        for exception in self.queue_manager.failure_descriptions():
            self.logger.info(exception)

    def data_loader_opts(self, working_dir: Path):
        """Callback for initializing a data loader task.

        For all of our dataloader sub-threads, these options
        will be applied.
        """
        wd = SnowfakeryWorkingDirectory(working_dir)

        options = {
            "mapping": wd.mapping_file,
            "reset_oids": False,
            "database_url": wd.database_url,
            "set_recently_viewed": False,
            "ignore_row_errors": self.ignore_row_errors,
            # don't need to pass loading_rules because they are merged into mapping
        }
        return options

    # TODO: This method is actually based on the number generated,
    #       because it is called before the load.
    #       If there are row errors, it will drift out of correctness
    #       Code needs to be updated to rename again after load.
    #       Or move away from using directory names for math altogether.
    def data_loader_new_directory_name(self, working_dir: Path):
        """Change the directory name to reflect the true number of sets created."""

        wd = SnowfakeryWorkingDirectory(working_dir)
        key = wd.index
        if key not in self.cached_counts:
            self.cached_counts[key] = wd.get_record_counts()

        if not self.run_until.sobject_name:
            return working_dir

        count = self.cached_counts[key][self.run_until.sobject_name]

        path, _ = str(working_dir).rsplit("_", 1)
        new_working_dir = Path(path + "_" + str(count))
        return new_working_dir

    def generator_data_dir(self, idx, template_path, batch_size, parent_dir):
        """Create a new generator directory with a name based on index and batch_size"""
        assert batch_size > 0
        data_dir = parent_dir / (str(idx) + "_" + str(batch_size))
        shutil.copytree(template_path, data_dir)
        return data_dir

    def get_upload_status(
        self,
        batch_size,
    ):
        """Combine information from the different data sources into a single "report".

        Useful for debugging but also for making decisions about what to do next."""

        return self.queue_manager.get_upload_status(
            batch_size, self.sets_finished_while_generating_template
        )

    @contextmanager
    def workingdir_or_tempdir(self, working_directory: T.Optional[T.Union[Path, str]]):
        """Make a working directory or a temporary directory, as needed"""
        if working_directory:
            working_directory = Path(working_directory)
            working_directory.mkdir()
            self.logger.info(f"Working Directory {working_directory}")
            yield working_directory
        elif self.debug_mode:
            working_directory = Path(mkdtemp())
            self.logger.info(
                f"Due to debug mode, Working Directory {working_directory} will not be removed"
            )
            yield working_directory
        else:
            with TemporaryDirectory() as tempdir:
                yield Path(tempdir)

    def _generate_and_load_initial_batch(self, working_directory: Path):
        """Generate a single batch to set up all just_once (singleton) objects"""

        template_dir = Path(working_directory) / "template_1"
        template_dir.mkdir()
        # changes here should often be reflected in
        # data_generator_opts and data_loader_opts

        channel_decl = self.channel_configs[0]

        plugin_options = {
            "pid": "0",
            "big_ids": "True",
        }
        results = self._generate_and_load_batch(
            template_dir,
            channel_decl.org_config,
            {
                "generator_yaml": self.options.get("recipe"),
                "num_records": 1,  # smallest possible batch to get to parallelizing fast
                "num_records_tablename": self.run_until.sobject_name or COUNT_REPS,
                "loading_rules": self.loading_rules,
                "vars": channel_decl.merge_recipe_options(self.recipe_options),
                "plugin_options": plugin_options,
                "bulk_mode": self.bulk_mode,
            },
        )
        self.update_running_totals_from_load_step_results(results)

        # rename directory to reflect real number of sets created.
        wd = SnowfakeryWorkingDirectory(template_dir)
        if self.run_until.sobject_name:
            self.sets_finished_while_generating_template = wd.get_record_counts()[
                self.run_until.sobject_name
            ]
        else:
            self.sets_finished_while_generating_template = 1

        new_template_dir = data_loader_new_directory_name(template_dir, self.run_until)
        shutil.move(template_dir, new_template_dir)
        template_dir = new_template_dir

        # don't send data tables to child processes. All they
        # care about are ID->OID mappings
        wd = SnowfakeryWorkingDirectory(template_dir)
        self._cleanup_object_tables(*wd.setup_engine())

        return template_dir, wd.relevant_sobjects()

    def _generate_and_load_batch(self, tempdir, org_config, options) -> dict:
        """Before the "full" dataload starts we do a single batch to
        load singletons.
        """
        options = {
            **options,
            "working_directory": tempdir,
            "set_recently_viewed": False,
            "ignore_row_errors": self.ignore_row_errors,
        }
        subtask_config = TaskConfig({"options": options})
        subtask = GenerateAndLoadDataFromYaml(
            project_config=self.project_config,
            task_config=subtask_config,
            org_config=org_config,
            flow=self.flow,
            name=self.name,
            stepnum=self.stepnum,
        )
        subtask()
        return subtask.return_values["load_results"][0]

    def _cleanup_object_tables(self, engine, metadata):
        """Delete all tables that do not relate to id->OID mapping"""
        tables = metadata.tables
        tables_to_drop = [
            table
            for tablename, table in tables.items()
            if not tablename.endswith("sf_ids")
        ]
        if tables_to_drop:
            metadata.drop_all(tables=tables_to_drop)


class RunningTotals:
    """Keep track of # of row errors and successess"""

    errors: int = 0
    successes: int = 0

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.__dict__}>"


class RulesFileAndRules(T.NamedTuple):
    file: Path
    rules: T.List[ChannelDeclaration]


def _read_channel_declarations_from_file(
    loading_rules_file: Path,
) -> RulesFileAndRules:
    """Look in a .load.yml file for the channel declarations"""
    with loading_rules_file.open() as f:
        decls = SObjectRuleDeclarationFile.parse_from_yaml(f)
        channel_decls = decls.channel_declarations or []
        assert isinstance(channel_decls, list)
        return RulesFileAndRules(loading_rules_file, channel_decls)


def read_channel_declarations(
    recipe: Path, loading_rules_files: T.List[Path]
) -> T.List[ChannelDeclaration]:
    """Find all appropriate channel declarations.

    Some discovered through naming conventions, others
    through command lines options or YML config."""
    implicit_rules_file = infer_load_file_path(recipe)
    if implicit_rules_file and implicit_rules_file.exists():
        loading_rules_files = loading_rules_files + [implicit_rules_file]
    # uniqify without losing order
    loading_rules_files = list(dict.fromkeys(loading_rules_files))

    rules_lists = [
        _read_channel_declarations_from_file(file) for file in loading_rules_files
    ]
    rules_lists = [rules_list for rules_list in rules_lists if rules_list.rules]

    if len(rules_lists) > 1:
        files = ", ".join(
            [f"{rules_list.file}: {rules_list.rules}" for rules_list in rules_lists]
        )
        msg = f"Multiple channel declarations: {files}"
        raise TaskOptionsError(msg)
    elif len(rules_lists) == 1:
        return rules_lists[0].rules
    else:
        return []


class ChannelConfig(T.NamedTuple):
    """A channel represents a connection to Salesforce via a username.

    It can also have recipe options and other documented properties.
    https://github.com/SFDO-Tooling/Snowfakery/search?q=ChannelDeclaration
    """

    org_config: OrgConfig
    declaration: ChannelDeclaration = None

    def merge_recipe_options(self, task_recipe_options):
        """Merge the recipe options from the channel declaration with those from the task config"""
        channel_options = self.declaration.recipe_options or {}
        task_recipe_options = task_recipe_options or {}
        self.check_conflicting_options(channel_options, task_recipe_options)
        recipe_options = {
            **task_recipe_options,
            **channel_options,
        }
        return recipe_options

    @staticmethod
    def check_conflicting_options(channel_options, task_recipe_options):
        """Check that options do not conflict"""
        double_specified_options = set(task_recipe_options.keys()).intersection(
            set(channel_options.keys())
        )
        conflicting_options = [
            optname
            for optname in double_specified_options
            if task_recipe_options[optname] != channel_options[optname]
        ]
        if conflicting_options:
            raise TaskOptionsError(
                f"Recipe options cannot conflict: {conflicting_options}"
            )


def standard_channel_config(
    org_config: OrgConfig,
    recipe_options: dict,
    num_generators: int,
    num_loaders: int = None,
):
    """Default configuration for a single-channel data-load"""
    channel = ChannelDeclaration(
        user="Username not used in this context",
        recipe_options=recipe_options,
        num_generators=num_generators,
        num_loaders=num_loaders,
    )

    return ChannelConfig(org_config, channel)


def channel_configs_from_decls(
    channel_decls: T.List[ChannelDeclaration],
    keychain: BaseProjectKeychain,
):
    """Reify channel configs and look up orgconfig"""

    def config_from_decl(decl):
        return ChannelConfig(keychain.get_org(decl.user), decl)

    return [config_from_decl(decl) for decl in channel_decls]



from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.salesforce_api.metadata import ApiListMetadata, ApiRetrieveUnpackaged
from cumulusci.tasks.salesforce import BaseRetrieveMetadata
from cumulusci.utils import package_xml_from_dict

retrieve_reportsanddashboards_options = BaseRetrieveMetadata.task_options.copy()
retrieve_reportsanddashboards_options.update(
    {
        "report_folders": {
            "description": "A list of the report folders to retrieve reports.  Separate by commas for multiple folders."
        },
        "dashboard_folders": {
            "description": "A list of the dashboard folders to retrieve reports.  Separate by commas for multiple folders."
        },
        "api_version": {
            "description": "Override the API version used to list metadata"
        },
    }
)


class RetrieveReportsAndDashboards(BaseRetrieveMetadata):
    api_class = ApiRetrieveUnpackaged
    list_metadata_api_class = ApiListMetadata

    task_options = retrieve_reportsanddashboards_options

    def _init_options(self, kwargs):
        super(RetrieveReportsAndDashboards, self)._init_options(kwargs)
        if "api_version" not in self.options:
            self.options[
                "api_version"
            ] = self.project_config.project__package__api_version

    def _validate_options(self):
        super(RetrieveReportsAndDashboards, self)._validate_options()
        if (
            "report_folders" not in self.options
            and "dashboard_folders" not in self.options
        ):
            raise TaskOptionsError(
                "You must provide at least one folder name for either report_folders or dashboard_folders"
            )

    def _get_api(self):
        metadata = {}
        if "report_folders" in self.options:
            for folder in self.options["report_folders"]:
                api_reports = self.list_metadata_api_class(
                    self,
                    "Report",
                    metadata=metadata,
                    folder=folder,
                    as_of_version=self.options["api_version"],
                )
                metadata = api_reports()
        if "dashboard_folders" in self.options:
            for folder in self.options["dashboard_folders"]:
                api_dashboards = self.list_metadata_api_class(
                    self,
                    "Dashboard",
                    metadata=metadata,
                    folder=folder,
                    as_of_version=self.options["api_version"],
                )
                metadata = api_dashboards()

        items = {}
        if "Report" in metadata:
            items["Report"] = []
            items["Report"].extend(self.options["report_folders"])
            for report in metadata["Report"]:
                items["Report"].append(report["fullName"])
        if "Dashboard" in metadata:
            items["Dashboard"] = []
            items["Dashboard"].extend(self.options["dashboard_folders"])
            for dashboard in metadata["Dashboard"]:
                items["Dashboard"].append(dashboard["fullName"])

        api_version = self.options["api_version"]
        package_xml = package_xml_from_dict(items, api_version)
        return self.api_class(self, package_xml, api_version)



import factory

from cumulusci.tasks.bulkdata.factory_utils import Models, ModuleDataFactory


class GenerateDummyData(ModuleDataFactory):
    """Generate data based on test mapping.yml"""

    def make_records(self, num_records, factories, current_batch_num):
        assert num_records % 4 == 0, "Use a batch size divisible by 4"
        factories.create_batch("ContactFactory", num_records // 2)
        factories["ContactFactory"].create_batch(num_records // 4)
        factories.create_batch("AccountFactory", num_records // 4)


class AccountFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = Models.Account

    id = factory.Sequence(lambda i: i)
    Name = factory.Sequence(lambda i: "Account %d" % i)
    BillingStreet = "Baker St."


class ContactFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        model = Models.Contact
        exclude = ["account"]  # don't try to attach the account to the Contact directly

    id = factory.Sequence(lambda n: n + 1)
    account = factory.SubFactory(AccountFactory)  # create account automatically
    AccountId = factory.LazyAttribute(lambda o: o.account.id)
    FirstName = factory.Faker("first_name")
    LastName = factory.Faker("last_name")
    Email = factory.Faker("email", domain="example.com")
    MailingStreet = "Baker St."



import json
import os
from datetime import datetime
from unittest import mock

import pytest
import responses
from github3.exceptions import (
    AuthenticationFailed,
    ConnectionError,
    ForbiddenError,
    TransportError,
)
from github3.pulls import ShortPullRequest
from github3.repos.repo import Repository
from github3.session import AppInstallationTokenAuth
from requests.exceptions import RequestException, RetryError
from requests.models import Response

import cumulusci
from cumulusci.core import github
from cumulusci.core.exceptions import (
    DependencyLookupError,
    GithubApiError,
    GithubApiNotFoundError,
    GithubException,
)
from cumulusci.core.github import (
    SSO_WARNING,
    UNAUTHORIZED_WARNING,
    add_labels_to_pull_request,
    catch_common_github_auth_errors,
    check_github_sso_auth,
    create_gist,
    create_pull_request,
    format_github3_exception,
    get_commit,
    get_github_api,
    get_github_api_for_repo,
    get_oauth_device_flow_token,
    get_oauth_scopes,
    get_pull_requests_by_commit,
    get_pull_requests_by_head,
    get_pull_requests_with_base_branch,
    get_ref_for_tag,
    get_sso_disabled_orgs,
    get_tag_by_name,
    get_version_id_from_tag,
    is_label_on_pull_request,
    is_pull_request_merged,
    markdown_link_to_pr,
    validate_service,
    warn_oauth_restricted,
)
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.release_notes.tests.utils import MockUtil


class TestGithub(GithubApiTestMixin):
    @classmethod
    def teardown_method(cls):
        # clear cached repo -> installation mapping
        github.INSTALLATIONS.clear()

    @pytest.fixture
    def mock_util(self):
        return MockUtil("TestOwner", "TestRepo")

    @pytest.fixture
    def repo(self, gh_api):
        repo_json = self._get_expected_repo("TestOwner", "TestRepo")
        return Repository(repo_json, gh_api)

    def test_github_api_retries(self, mock_http_response):
        gh = get_github_api("TestUser", "TestPass")
        adapter = gh.session.get_adapter("http://")

        assert 0.3 == adapter.max_retries.backoff_factor
        assert 502 in adapter.max_retries.status_forcelist

        with mock.patch(
            "urllib3.connectionpool.HTTPConnectionPool._make_request"
        ) as _make_request:
            _make_request.side_effect = [
                mock_http_response(status=503),
                mock_http_response(status=200),
            ]

            gh.octocat("meow")
            assert 2 == _make_request.call_count

    @responses.activate
    @mock.patch("github3.apps.create_token")
    def test_get_github_api_for_repo(self, create_token):
        create_token.return_value = "ATOKEN"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/installation",
            json={
                "id": 1,
                "access_tokens_url": "",
                "account": "",
                "app_id": "",
                "created_at": "",
                "events": "",
                "html_url": "",
                "permissions": "",
                "repositories_url": "",
                "repository_selection": "",
                "single_file_name": "",
                "target_id": "",
                "target_type": "",
                "updated_at": "",
            },
        )
        responses.add(
            "POST",
            "https://api.github.com/app/installations/1/access_tokens",
            status=201,
            json={"token": "ITOKEN", "expires_at": datetime.now().isoformat()},
        )

        with mock.patch.dict(
            os.environ, {"GITHUB_APP_KEY": "bogus", "GITHUB_APP_ID": "1234"}
        ):
            gh = get_github_api_for_repo(None, "TestOwner", "TestRepo")
            assert isinstance(gh.session.auth, AppInstallationTokenAuth)

    @responses.activate
    @mock.patch("github3.apps.create_token")
    def test_get_github_api_for_repo__not_installed(self, create_token):
        create_token.return_value = "ATOKEN"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/installation",
            status=404,
        )
        with mock.patch.dict(
            os.environ, {"GITHUB_APP_KEY": "bogus", "GITHUB_APP_ID": "1234"}
        ):
            with pytest.raises(GithubException):
                get_github_api_for_repo(None, "TestOwner", "TestRepo")

    @responses.activate
    @mock.patch("cumulusci.core.github.GitHub")
    def test_get_github_api_for_repo__token(self, GitHub):
        with mock.patch.dict(os.environ, {"GITHUB_TOKEN": "token"}):
            gh = get_github_api_for_repo(None, "TestOwner", "TestRepo")
        gh.login.assert_called_once_with(token="token")

    @responses.activate
    def test_validate_service(self):
        responses.add("GET", "https://api.github.com/user", status=401, headers={})
        with pytest.raises(GithubException):
            validate_service({"username": "BOGUS", "token": "BOGUS"})

    @responses.activate
    def test_get_pull_requests_by_head(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(
            pulls=self._get_expected_pull_requests(1),
            head=repo.owner.login + ":" + "some-other-branch",
        )
        pull_requests = get_pull_requests_by_head(repo, "some-other-branch")
        assert 1 == len(pull_requests)

        # ConnectionError present when we reachout with
        # a branch name (url parameter) that we aren't expecting
        with pytest.raises(ConnectionError):
            get_pull_requests_by_head(repo, "does-not-exist")

    @responses.activate
    def test_get_pull_requests_by_head__no_pulls(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls()
        pull_requests = get_pull_requests_by_head(repo, "test_branch")
        assert pull_requests == []

        pull_requests = get_pull_requests_by_head(repo, "main")
        assert pull_requests is None

    @responses.activate
    def test_get_pull_request_by_head__multiple_pulls(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(pulls=self._get_expected_pull_requests(2))
        pull_requests = get_pull_requests_by_head(repo, "test_branch")
        assert 2 == len(pull_requests)

    @responses.activate
    def test_get_pull_requests_with_base_branch(self, mock_util, repo):
        self.init_github()
        mock_util.mock_pulls(base="main", head="TestOwner:some-branch")
        pull_requests = get_pull_requests_with_base_branch(
            repo, "main", head="some-branch"
        )
        assert 0 == len(pull_requests)

        responses.reset()
        mock_util.mock_pulls(pulls=self._get_expected_pull_requests(3), base="main")
        pull_requests = get_pull_requests_with_base_branch(repo, "main")
        assert 3 == len(pull_requests)

    @responses.activate
    def test_create_pull_request(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.mock_pulls(
            method=responses.POST,
            pulls=self._get_expected_pull_request(1, 1, "Test Body"),
        )
        pull_request = create_pull_request(repo, "test-branch")
        assert pull_request is not None
        assert pull_request.body == "Test Body"

    @responses.activate
    def test_is_label_on_pr(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.add_issue_response(self._get_expected_issue(1))
        mock_util.add_issue_response(self._get_expected_issue(2))
        mock_util.mock_issue_labels(1, responses.GET, ["Octocat", "bogus 1", "bogus 2"])
        mock_util.mock_issue_labels(2, responses.GET, ["bogus 1", "bogus 2"])

        pull_request = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        pull_request.number = 1
        assert is_label_on_pull_request(repo, pull_request, "Octocat")
        pull_request.number = 2
        assert not is_label_on_pull_request(repo, pull_request, "Octocat")

    @responses.activate
    def test_add_labels_to_pull_request(self, mock_util, repo, gh_api):
        self.init_github()
        mock_util.add_issue_response(self._get_expected_issue(1))
        mock_util.mock_issue_labels(1, method=responses.POST)
        pull_request = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        pull_request.number = 1

        add_labels_to_pull_request(repo, pull_request, "first", "second", "third")
        body = responses.calls[-1].request.body
        assert "first" in body
        assert "second" in body
        assert "third" in body

    @responses.activate
    def test_get_pull_request_by_commit(self, mock_util, repo, gh_api):
        self.init_github()
        commit_sha = "asdf1234asdf1234"
        mock_util.mock_pull_request_by_commit_sha(commit_sha)
        pull_requests = get_pull_requests_by_commit(gh_api, repo, commit_sha)
        assert len(pull_requests) == 1

    def test_is_pull_request_merged(self, gh_api):
        self.init_github()

        merged_pull_request = ShortPullRequest(
            self._get_expected_pull_request(1, 1), gh_api
        )
        merged_pull_request.merged_at = "DateTimeStr"

        unmerged_pull_request = ShortPullRequest(
            self._get_expected_pull_request(1, 1), gh_api
        )
        unmerged_pull_request.merged_at = None

        assert is_pull_request_merged(merged_pull_request)
        assert not is_pull_request_merged(unmerged_pull_request)

    def test_markdown_link_to_pr(self, gh_api):
        self.init_github()
        pr = ShortPullRequest(self._get_expected_pull_request(1, 1), gh_api)
        actual_link = markdown_link_to_pr(pr)
        expected_link = f"{pr.title} [[PR{pr.number}]({pr.html_url})]"

        assert expected_link == actual_link

    @responses.activate
    def test_create_gist(self, gh_api, mock_util):
        self.init_github()

        description = "Test Gist Creation"
        filename = "error_output.txt"
        content = "Hello there gist!"
        files = {filename: content}

        self.mock_gist(description, files)
        gist = create_gist(gh_api, description, files)

        expected_url = f"https://gist.github.com/{gist.id}"
        assert expected_url == gist.html_url

    def mock_gist(self, description, files):
        responses.add(
            method=responses.POST,
            url="https://api.github.com/gists",
            json=self._get_expected_gist(description, files),
            status=201,
        )

    @responses.activate
    def test_get_tag_by_name(self, repo):
        self.init_github()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA"),
            status=200,
        )
        tag = get_tag_by_name(repo, "tag_SHA")
        assert tag.tag == "beta/1.0"

    @responses.activate
    def test_get_tag_by_name__404(self, repo):
        self.init_github()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA"),
            status=404,
        )
        with pytest.raises(GithubApiNotFoundError):
            get_tag_by_name(repo, "tag_SHA")

    @responses.activate
    def test_current_tag_is_lightweight(self, repo):
        self.init_github()
        light_tag = self._get_expected_tag_ref("tag_SHA", "tag_SHA")
        light_tag["object"]["type"] = "commit"
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=light_tag,
            status=200,
        )
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA"),
            status=404,
        )
        with pytest.raises(GithubApiNotFoundError) as exc:
            get_tag_by_name(repo, "tag_SHA")

        assert "not an annotated tag" in str(exc)

    @responses.activate
    def test_get_ref_by_name(self, repo):
        self.init_github()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=200,
        )
        ref = get_ref_for_tag(repo, "tag_SHA")
        assert ref.object.sha == "tag_SHA"

    @responses.activate
    def test_get_ref_by_name__404(self, repo):
        self.init_github()
        responses.add(
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/tag_SHA",
            json=self._get_expected_tag_ref("tag_SHA", "tag_SHA"),
            status=404,
        )
        with pytest.raises(GithubApiNotFoundError):
            get_ref_for_tag(repo, "tag_SHA")

    @responses.activate
    def test_get_version_id_from_tag(self, repo):
        self.init_github()
        responses.add(  # the ref for the tag is fetched first
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/test-tag-name",
            json=self._get_expected_tag_ref("test-tag-name", "tag_SHA"),
            status=200,
        )
        tag_message = """Release of Test Package\nversion_id: 04t000000000000\n\ndependencies: []"""
        responses.add(  # then we fetch that actual tag with the ref
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA", message=tag_message),
            status=200,
        )
        version_id = get_version_id_from_tag(repo, "test-tag-name")
        assert version_id == "04t000000000000"

    @responses.activate
    def test_get_version_id_from_tag__dependency_error(self, repo):
        self.init_github()
        responses.add(  # the ref for the tag is fetched first
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/ref/tags/test-tag-name",
            json=self._get_expected_tag_ref("test-tag-name", "tag_SHA"),
            status=200,
        )
        tag_message = (
            """Release of Test Package\nversion_id: invalid_id\n\ndependencies: []"""
        )
        responses.add(  # then we fetch that actual tag with the ref
            "GET",
            "https://api.github.com/repos/TestOwner/TestRepo/git/tags/tag_SHA",
            json=self._get_expected_tag("beta/1.0", "tag_SHA", message=tag_message),
            status=200,
        )
        with pytest.raises(DependencyLookupError):
            get_version_id_from_tag(repo, "test-tag-name")

    @responses.activate
    def test_get_commit(self, repo):
        self.init_github()
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/commits/DUMMY_SHA",
            json=self._get_expected_commit("DUMMY_SHA"),
            status=200,
        )
        commit = get_commit(repo, "DUMMY_SHA")
        assert commit.sha == "DUMMY_SHA"

    @responses.activate
    def test_get_commit__dependency_error(self, repo):
        self.init_github()
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/commits/DUMMY_SHA",
            json={"message": "Could not verify object"},
            status=422,
        )
        with pytest.raises(
            DependencyLookupError, match="Could not find commit DUMMY_SHA on GitHub"
        ):
            get_commit(repo, "DUMMY_SHA")

    def test_get_oauth_scopes(self):
        resp = Response()
        resp.headers["X-OAuth-Scopes"] = "repo, user"
        assert {"repo", "user"} == get_oauth_scopes(resp)

        resp = Response()
        resp.headers["X-OAuth-Scopes"] = "repo"
        assert {"repo"} == get_oauth_scopes(resp)

        resp = Response()
        assert set() == get_oauth_scopes(resp)

    def test_get_sso_disabled_orgs(self):
        resp = Response()
        assert [] == get_sso_disabled_orgs(resp)

        resp = Response()
        resp.headers["X-Github-Sso"] = "partial-results; organizations=0810298,20348880"
        assert ["0810298", "20348880"] == get_sso_disabled_orgs(resp)

    def test_check_github_sso_no_forbidden(self):
        resp = Response()
        resp.status_code = 401
        exc = AuthenticationFailed(resp)
        assert "" == check_github_sso_auth(exc)

        resp.status_code = 403
        exc = ForbiddenError(resp)
        assert "" == check_github_sso_auth(exc)

    @mock.patch("webbrowser.open")
    def test_check_github_sso_unauthorized_token(self, browser_open):
        resp = Response()
        resp.status_code = 403
        auth_url = "https://github.com/orgs/foo/sso?authorization_request=longhash"
        resp.headers["X-Github-Sso"] = f"required; url={auth_url}"
        exc = ForbiddenError(resp)

        check_github_sso_auth(exc)

        browser_open.assert_called_with(auth_url)

    def test_check_github_sso_partial_auth(self):
        resp = Response()
        resp.status_code = 403
        resp.headers["X-Github-Sso"] = "partial-results; organizations=0810298,20348880"
        exc = ForbiddenError(resp)

        expected_err_msg = f"{SSO_WARNING} ['0810298', '20348880']"
        actual_error_msg = check_github_sso_auth(exc).strip()
        assert expected_err_msg == actual_error_msg

    def test_github_oauth_org_restricted(self):
        resp = Response()
        resp.status_code = 403
        body = {"message": "organization has enabled OAuth App access restrictions"}
        resp._content = json.dumps(body)
        exc = ForbiddenError(resp)

        expected_warning = "You may also use a Personal Access Token as a workaround."
        actual_error_msg = warn_oauth_restricted(exc)
        assert expected_warning in actual_error_msg

    def test_format_gh3_exc_retry(self):
        resp = Response()
        resp.status_code = 401
        message = "Max retries exceeded with url: foo (Caused by ResponseError('too many 401 error responses',))"
        base_exc = RetryError(message, response=resp)
        exc = TransportError(base_exc)
        assert UNAUTHORIZED_WARNING == format_github3_exception(exc)

    def test_catch_common_decorator(self):
        resp = Response()
        resp.status_code = 403
        resp.headers["X-Github-Sso"] = "partial-results; organizations=0810298,20348880"

        expected_err_msg = "Results may be incomplete. You have not granted your Personal Access token access to the following organizations: ['0810298', '20348880']"

        @catch_common_github_auth_errors
        def test_func():
            raise ForbiddenError(resp)

        with pytest.raises(GithubApiError) as exc:
            test_func()
            actual_error_msg = exc.message
            assert expected_err_msg == actual_error_msg

    def test_catch_common_decorator_ignores(self):
        resp = Response()
        resp.status_code = 401

        @catch_common_github_auth_errors
        def test_func():
            e = RequestException(response=resp)
            raise TransportError(e)

        with pytest.raises(TransportError):
            test_func()

    @responses.activate
    def test_validate_no_repo_exc(self):
        service_dict = {
            "username": "e2ac67",
            "token": "ghp_cf83e1357eefb8bdf1542850d66d8007d620e4",
            "email": "testerson@test.com",
        }
        responses.add(
            responses.GET,
            "https://api.github.com/user",
            json={
                "login": "e2ac67",
                "id": 91303375,
                "node_id": "MDQ6VXNlcjkxMzAzMzc1",
                "avatar_url": "https://avatars.githubusercontent.com/u/91303375?v=4",
                "gravatar_id": "",
                "url": "https://api.github.com/users/e2ac67",
                "html_url": "https://github.com/e2ac67",
                "followers_url": "https://api.github.com/users/e2ac67/followers",
                "following_url": "https://api.github.com/users/e2ac67/following{/other_user}",
                "gists_url": "https://api.github.com/users/e2ac67/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/e2ac67/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/e2ac67/subscriptions",
                "organizations_url": "https://api.github.com/users/e2ac67/orgs",
                "repos_url": "https://api.github.com/users/e2ac67/repos",
                "events_url": "https://api.github.com/users/e2ac67/events{/privacy}",
                "received_events_url": "https://api.github.com/users/e2ac67/received_events",
                "type": "User",
                "site_admin": False,
                "name": None,
                "company": None,
                "blog": "",
                "location": None,
                "email": None,
                "hireable": None,
                "bio": None,
                "twitter_username": None,
                "public_repos": 0,
                "public_gists": 0,
                "followers": 0,
                "following": 0,
                "created_at": "2021-09-24T03:53:02Z",
                "updated_at": "2021-09-24T03:59:40Z",
            },
        )
        responses.add(
            responses.GET,
            "https://api.github.com/user/orgs",
            json=[],
        )
        responses.add(
            responses.GET,
            "https://api.github.com/user/repos",
            json=[],
            headers={
                "GitHub-Authentication-Token-Expiration": "2021-10-07 19:07:53 UTC",
                "X-OAuth-Scopes": "gist, repo",
            },
        )
        updated_dict = validate_service(service_dict)
        expected_dict = {
            "username": "e2ac67",
            "token": "ghp_cf83e1357eefb8bdf1542850d66d8007d620e4",
            "email": "testerson@test.com",
            "Organizations": "",
            "scopes": {"gist", "repo"},
            "expires": "2021-10-07 19:07:53 UTC",
        }
        assert expected_dict == updated_dict

    @responses.activate
    def test_validate_bad_auth(self):
        service_dict = {
            "username": "e2ac67",
            "token": "bad_cf83e1357eefb8bdf1542850d66d8007d620e4",
            "email": "testerson@test.com",
        }

        responses.add(
            responses.GET,
            "https://api.github.com/user",
            json={
                "message": "Bad credentials",
                "documentation_url": "https://docs.github.com/rest",
            },
            status=401,
        )

        with pytest.raises(cumulusci.core.exceptions.GithubException) as e:
            validate_service(service_dict)
        assert "401" in str(e.value)

    @mock.patch("webbrowser.open")
    @mock.patch("cumulusci.core.github.get_device_code")
    @mock.patch("cumulusci.core.github.get_device_oauth_token", autospec=True)
    def test_get_oauth_device_flow_token(
        self,
        get_token,
        get_code,
        browser_open,
    ):
        device_config = {
            "device_code": "36482450e39b7f27d9a145a96898d29365a4e73f",
            "user_code": "3E15-9D06",
            "verification_uri": "https://github.com/login/device",
            "expires_in": 899,
            "interval": 5,
        }
        get_code.return_value = device_config
        get_token.return_value = {
            "access_token": "expected_access_token",
            "token_type": "bearer",
            "scope": "gist,repo",
        }

        returned_token = get_oauth_device_flow_token()

        assert "expected_access_token" == returned_token
        get_token.assert_called_once()
        get_code.assert_called_once()
        browser_open.assert_called_with("https://github.com/login/device")



import http.client
import logging
import os
import shutil
import tempfile
import unittest
from copy import deepcopy
from distutils.version import StrictVersion
from unittest.mock import MagicMock, Mock, patch

import pytest
import responses
from simple_salesforce import SalesforceGeneralError

from cumulusci.core import exceptions as exc
from cumulusci.core.config import (
    BaseProjectConfig,
    OrgConfig,
    TaskConfig,
    UniversalConfig,
)
from cumulusci.core.exceptions import (
    ApexCompilationException,
    ApexException,
    ApexTestException,
    CumulusCIException,
    SalesforceException,
    TaskOptionsError,
)
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.tests.utils import MockLoggerMixin
from cumulusci.tasks.apex.anon import AnonymousApexTask
from cumulusci.tasks.apex.batch import BatchApexWait
from cumulusci.tasks.apex.testrunner import RunApexTests


@patch(
    "cumulusci.tasks.salesforce.BaseSalesforceTask._update_credentials",
    MagicMock(return_value=None),
)
class TestRunApexTests(MockLoggerMixin, unittest.TestCase):
    def setUp(self):
        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages
        self.api_version = 38.0
        self.universal_config = UniversalConfig(
            {"project": {"api_version": self.api_version}}
        )
        self.task_config = TaskConfig()
        self.task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
        }
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )
        self.project_config.config["project"] = {
            "package": {"api_version": self.api_version}
        }
        keychain = BaseProjectKeychain(self.project_config, "")
        self.project_config.set_keychain(keychain)
        self.org_config = OrgConfig(
            {
                "id": "foo/1",
                "instance_url": "https://example.com",
                "access_token": "abc123",
            },
            "test",
        )
        self.base_tooling_url = "{}/services/data/v{}/tooling/".format(
            self.org_config.instance_url, self.api_version
        )

    def _mock_apex_class_query(self, name="TestClass_TEST", namespace=None):
        namespace_param = "null" if namespace is None else f"%27{namespace}%27"
        url = (
            self.base_tooling_url
            + "query/?q=SELECT+Id%2C+Name+"
            + f"FROM+ApexClass+WHERE+NamespacePrefix+%3D+{namespace_param}"
            + "+AND+%28Name+LIKE+%27%25_TEST%27%29"
        )
        expected_response = {
            "done": True,
            "records": [{"Id": 1, "Name": name}],
            "totalSize": 1,
        }
        responses.add(
            responses.GET, url, match_querystring=True, json=expected_response
        )

    def _get_mock_test_query_results(self, methodnames, outcomes, messages):
        record_base = {
            "attributes": {
                "type": "ApexTestResult",
                "url": "/services/data/v40.0/tooling/sobjects/ApexTestResult/07M41000009gbT3EAI",
            },
            "ApexClass": {
                "attributes": {
                    "type": "ApexClass",
                    "url": "/services/data/v40.0/tooling/sobjects/ApexClass/01p4100000Fu4Z0AAJ",
                },
                "Name": "EP_TaskDependency_TEST",
            },
            "ApexClassId": 1,
            "ApexLogId": 1,
            "TestTimestamp": "2017-07-18T20:36:04.000+0000",
            "Id": "07M41000009gbT3EAI",
            "Message": None,
            "MethodName": None,
            "Outcome": None,
            "QueueItem": {
                "attributes": {
                    "type": "ApexTestQueueItem",
                    "url": "/services/data/v40.0/tooling/sobjects/ApexTestQueueItem/70941000000q7VsAAI",
                },
                "Status": "Completed",
                "ExtendedStatus": "(4/4)",
            },
            "RunTime": 1707,
            "StackTrace": "1. ParentFunction\n2. ChildFunction",
            "Status": "Completed",
            "Name": "TestClass_TEST",
            "ApexTestResults": {
                "size": 1,
                "totalSize": 1,
                "done": True,
                "queryLocator": None,
                "entityTypeName": "ApexTestResultLimits",
                "records": [
                    {
                        "attributes": {
                            "type": "ApexTestResultLimits",
                            "url": "/services/data/v40.0/tooling/sobjects/ApexTestResultLimits/05n41000002Y7OQAA0",
                        },
                        "Id": "05n41000002Y7OQAA0",
                        "Callouts": 0,
                        "AsyncCalls": 0,
                        "DmlRows": 5,
                        "Email": 0,
                        "LimitContext": "SYNC",
                        "LimitExceptions": None,
                        "MobilePush": 0,
                        "QueryRows": 20,
                        "Sosl": 0,
                        "Cpu": 471,
                        "Dml": 4,
                        "Soql": 5,
                    }
                ],
            },
        }

        return_value = {"done": True, "records": []}

        for (method_name, outcome, message) in zip(methodnames, outcomes, messages):
            this_result = deepcopy(record_base)
            this_result["Message"] = message
            this_result["Outcome"] = outcome
            this_result["MethodName"] = method_name
            return_value["records"].append(this_result)

        return return_value

    def _get_mock_test_query_url(self, job_id):
        return (
            self.base_tooling_url
            + "query/?q=%0ASELECT+Id%2CApexClassId%2CTestTimestamp%2C%0A+++++++Message%2CMethodName%2COutcome%2C%0A+++++++RunTime%2CStackTrace%2C%0A+++++++%28SELECT%0A++++++++++Id%2CCallouts%2CAsyncCalls%2CDmlRows%2CEmail%2C%0A++++++++++LimitContext%2CLimitExceptions%2CMobilePush%2C%0A++++++++++QueryRows%2CSosl%2CCpu%2CDml%2CSoql%0A++++++++FROM+ApexTestResults%29%0AFROM+ApexTestResult%0AWHERE+AsyncApexJobId%3D%27{}%27%0A".format(
                job_id
            )
        )

    def _get_mock_testqueueitem_status_query_url(self, job_id):
        return (
            self.base_tooling_url
            + f"query/?q=SELECT+Id%2C+Status%2C+ExtendedStatus%2C+ApexClassId+FROM+ApexTestQueueItem+WHERE+ParentJobId+%3D+%27{job_id}%27+AND+Status+%3D+%27Failed%27"
        )

    def _mock_get_test_results(
        self, outcome="Pass", message="Test Passed", job_id="JOB_ID1234567"
    ):
        url = self._get_mock_test_query_url(job_id)

        expected_response = self._get_mock_test_query_results(
            ["TestMethod"], [outcome], [message]
        )
        responses.add(
            responses.GET, url, match_querystring=True, json=expected_response
        )

    def _mock_get_test_results_multiple(
        self, method_names, outcomes, messages, job_id="JOB_ID1234567"
    ):
        url = self._get_mock_test_query_url(job_id)

        expected_response = self._get_mock_test_query_results(
            method_names, outcomes, messages
        )
        responses.add(
            responses.GET, url, match_querystring=True, json=expected_response
        )

    def _mock_get_failed_test_classes(self, job_id="JOB_ID1234567"):
        url = self._get_mock_testqueueitem_status_query_url(job_id)

        responses.add(
            responses.GET,
            url,
            match_querystring=True,
            json={"totalSize": 0, "records": [], "done": True},
        )

    def _mock_get_failed_test_classes_failure(self, job_id="JOB_ID1234567"):
        url = self._get_mock_testqueueitem_status_query_url(job_id)

        responses.add(
            responses.GET,
            url,
            match_querystring=True,
            json={
                "totalSize": 1,
                "records": [
                    {
                        "Id": "0000000000000000",
                        "ApexClassId": 1,
                        "Status": "Failed",
                        "ExtendedStatus": "Double-plus ungood",
                    }
                ],
                "done": True,
            },
        )

    def _mock_get_symboltable(self):
        url = (
            self.base_tooling_url
            + "query/?q=SELECT+SymbolTable+FROM+ApexClass+WHERE+Name%3D%27TestClass_TEST%27"
        )

        responses.add(
            responses.GET,
            url,
            json={
                "records": [
                    {
                        "SymbolTable": {
                            "methods": [
                                {"name": "test1", "annotations": [{"name": "isTest"}]}
                            ]
                        }
                    }
                ]
            },
        )

    def _mock_get_symboltable_failure(self):
        url = (
            self.base_tooling_url
            + "query/?q=SELECT+SymbolTable+FROM+ApexClass+WHERE+Name%3D%27TestClass_TEST%27"
        )

        responses.add(responses.GET, url, json={"records": []})

    def _mock_tests_complete(self, job_id="JOB_ID1234567"):
        url = (
            self.base_tooling_url
            + "query/?q=SELECT+Id%2C+Status%2C+"
            + "ApexClassId+FROM+ApexTestQueueItem+WHERE+ParentJobId+%3D+%27"
            + "{}%27".format(job_id)
        )
        expected_response = {
            "done": True,
            "totalSize": 1,
            "records": [{"Status": "Completed"}],
        }
        responses.add(
            responses.GET, url, match_querystring=True, json=expected_response
        )

    def _mock_tests_processing(self, job_id="JOB_ID1234567"):
        url = (
            self.base_tooling_url
            + "query/?q=SELECT+Id%2C+Status%2C+"
            + "ApexClassId+FROM+ApexTestQueueItem+WHERE+ParentJobId+%3D+%27"
            + "{}%27".format(job_id)
        )
        expected_response = {
            "done": True,
            "totalSize": 1,
            "records": [{"Status": "Processing", "ApexClassId": 1}],
        }
        responses.add(
            responses.GET, url, match_querystring=True, json=expected_response
        )

    def _mock_run_tests(self, success=True, body="JOB_ID1234567"):
        url = self.base_tooling_url + "runTestsAsynchronous"
        if success:
            responses.add(responses.POST, url, json=body)
        else:
            responses.add(responses.POST, url, status=http.client.SERVICE_UNAVAILABLE)

    @responses.activate
    def test_run_task(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_get_failed_test_classes()
        self._mock_tests_complete()
        self._mock_get_test_results()
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        task()
        self.assertEqual(len(responses.calls), 5)

    @responses.activate
    def test_run_task__server_error(self):
        self._mock_apex_class_query()
        self._mock_run_tests(success=False)
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        with self.assertRaises(SalesforceGeneralError):
            task()

    @responses.activate
    def test_run_task__failed(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_get_failed_test_classes()
        self._mock_tests_complete()
        self._mock_get_test_results("Fail")
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        with self.assertRaises(ApexTestException):
            task()

    @responses.activate
    def test_run_task__failed_class_level(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_get_failed_test_classes_failure()
        self._mock_tests_complete()
        self._mock_get_test_results()
        self._mock_get_symboltable()
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        with self.assertRaises(ApexTestException):
            task()

    @responses.activate
    def test_run_task__failed_class_level_no_symboltable(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_get_failed_test_classes_failure()
        self._mock_tests_complete()
        self._mock_get_test_results()
        self._mock_get_symboltable_failure()
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        with self.assertRaises(CumulusCIException):
            task()

    @responses.activate
    def test_run_task__failed_class_level_no_symboltable__spring20_managed(self):
        self._mock_apex_class_query(name="ns__Test_TEST", namespace="ns")
        self._mock_run_tests()
        self._mock_get_failed_test_classes_failure()
        self._mock_tests_complete()
        self._mock_get_test_results()
        self._mock_get_symboltable_failure()
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "managed": True,
            "namespace": "ns",
        }

        task = RunApexTests(self.project_config, task_config, self.org_config)
        task._get_test_methods_for_class = Mock()

        task()

        task._get_test_methods_for_class.assert_not_called()

    @responses.activate
    def test_run_task__retry_tests(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_run_tests(body="JOBID_9999")
        self._mock_get_failed_test_classes()
        self._mock_get_failed_test_classes(job_id="JOBID_9999")
        self._mock_tests_complete()
        self._mock_tests_complete(job_id="JOBID_9999")
        self._mock_get_test_results("Fail", "UNABLE_TO_LOCK_ROW")
        self._mock_get_test_results(job_id="JOBID_9999")

        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "retry_failures": ["UNABLE_TO_LOCK_ROW"],
        }
        task = RunApexTests(self.project_config, task_config, self.org_config)
        task()
        self.assertEqual(len(responses.calls), 9)

    @responses.activate
    def test_run_task__retry_tests_with_retry_always(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_run_tests(body="JOBID_9999")
        self._mock_run_tests(body="JOBID_9990")
        self._mock_get_failed_test_classes()
        self._mock_get_failed_test_classes(job_id="JOBID_9999")
        self._mock_get_failed_test_classes(job_id="JOBID_9990")
        self._mock_tests_complete()
        self._mock_tests_complete(job_id="JOBID_9999")
        self._mock_tests_complete(job_id="JOBID_9990")
        self._mock_get_test_results_multiple(
            ["TestOne", "TestTwo"],
            ["Fail", "Fail"],
            ["UNABLE_TO_LOCK_ROW", "LimitException"],
        )
        self._mock_get_test_results_multiple(
            ["TestOne"], ["Pass"], [""], job_id="JOBID_9999"
        )
        self._mock_get_test_results_multiple(
            ["TestTwo"], ["Fail"], ["LimitException"], job_id="JOBID_9990"
        )
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "retry_failures": ["UNABLE_TO_LOCK_ROW"],
            "retry_always": True,
        }
        task = RunApexTests(self.project_config, task_config, self.org_config)
        with self.assertRaises(ApexTestException):
            task()

    @responses.activate
    def test_run_task__retry_tests_fails(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_run_tests(body="JOBID_9999")
        self._mock_get_failed_test_classes()
        self._mock_get_failed_test_classes(job_id="JOBID_9999")
        self._mock_tests_complete()
        self._mock_tests_complete(job_id="JOBID_9999")
        self._mock_get_test_results("Fail", "UNABLE_TO_LOCK_ROW")
        self._mock_get_test_results("Fail", "DUPLICATES_DETECTED", job_id="JOBID_9999")

        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "retry_failures": ["UNABLE_TO_LOCK_ROW"],
        }
        task = RunApexTests(self.project_config, task_config, self.org_config)
        with self.assertRaises(ApexTestException):
            task()

    @responses.activate
    def test_run_task__processing(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_tests_processing()
        self._mock_get_failed_test_classes()
        self._mock_tests_complete()
        self._mock_get_test_results()
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        task()
        log = self._task_log_handler.messages
        assert "Completed: 0  Processing: 1 (TestClass_TEST)  Queued: 0" in log["info"]

    @responses.activate
    def test_run_task__not_verbose(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_tests_processing()
        self._mock_get_failed_test_classes()  # this returns all passes
        self._mock_tests_complete()
        self._mock_get_test_results()
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        task()
        log = self._task_log_handler.messages
        assert "Class: TestClass_TEST" not in log["info"]

    @responses.activate
    def test_run_task__verbose(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_get_failed_test_classes_failure()
        self._mock_tests_complete()
        self._mock_get_test_results()
        self._mock_get_symboltable()
        task_config = TaskConfig()
        task_config.config["options"] = {
            "verbose": True,
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
        }
        task = RunApexTests(self.project_config, task_config, self.org_config)
        with self.assertRaises(CumulusCIException):
            task()
        log = self._task_log_handler.messages
        assert "Class: TestClass_TEST" in log["info"]

    @responses.activate
    def test_run_task__no_code_coverage(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_get_failed_test_classes()
        self._mock_tests_complete()
        self._mock_get_test_results()
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
        }
        task = RunApexTests(self.project_config, task_config, self.org_config)
        task._check_code_coverage = Mock()
        task()
        task._check_code_coverage.assert_not_called()

    @responses.activate
    def test_run_task__checks_code_coverage(self):
        self._mock_apex_class_query()
        self._mock_run_tests()
        self._mock_get_failed_test_classes()
        self._mock_tests_complete()
        self._mock_get_test_results()
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "required_org_code_coverage_percent": "90",
        }

        org_config = OrgConfig(
            {
                "id": "foo/1",
                "instance_url": "https://example.com",
                "access_token": "abc123",
            },
            "test",
        )
        org_config._installed_packages = {"TEST": StrictVersion("1.2.3")}
        task = RunApexTests(self.project_config, task_config, org_config)
        task._check_code_coverage = Mock()
        task()
        task._check_code_coverage.assert_called_once()

    def test_code_coverage_integer(self):
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "required_org_code_coverage_percent": 90,
        }

        org_config = OrgConfig(
            {
                "id": "foo/1",
                "instance_url": "https://example.com",
                "access_token": "abc123",
            },
            "test",
        )
        task = RunApexTests(self.project_config, task_config, org_config)

        assert task.code_coverage_level == 90

    def test_code_coverage_percentage(self):
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "required_org_code_coverage_percent": "90%",
        }

        org_config = OrgConfig(
            {
                "id": "foo/1",
                "instance_url": "https://example.com",
                "access_token": "abc123",
            },
            "test",
        )
        task = RunApexTests(self.project_config, task_config, org_config)

        assert task.code_coverage_level == 90

    def test_exception_bad_code_coverage(self):
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "required_org_code_coverage_percent": "foo",
        }

        with self.assertRaises(TaskOptionsError):
            RunApexTests(self.project_config, task_config, self.org_config)

    @responses.activate
    def test_run_task__code_coverage_managed(self):
        self._mock_apex_class_query(namespace="TEST")
        self._mock_run_tests()
        self._mock_get_failed_test_classes()
        self._mock_tests_complete()
        self._mock_get_test_results()
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "namespace": "TEST",
            "required_org_code_coverage_percent": "90",
        }
        org_config = OrgConfig(
            {
                "id": "foo/1",
                "instance_url": "https://example.com",
                "access_token": "abc123",
            },
            "test",
        )
        org_config._installed_packages = {"TEST": StrictVersion("1.2.3")}

        task = RunApexTests(self.project_config, task_config, org_config)
        task._check_code_coverage = Mock()
        task()
        task._check_code_coverage.assert_not_called()

    def test_check_code_coverage(self):
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        task.code_coverage_level = 90
        task.tooling = Mock()
        task.tooling.query.return_value = {
            "records": [{"PercentCovered": 90}],
            "totalSize": 1,
        }

        task._check_code_coverage()
        task.tooling.query.assert_called_once_with(
            "SELECT PercentCovered FROM ApexOrgWideCoverage"
        )

    def test_check_code_coverage__fail(self):
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        task.code_coverage_level = 90
        task.tooling = Mock()
        task.tooling.query.return_value = {
            "records": [{"PercentCovered": 89}],
            "totalSize": 1,
        }

        with self.assertRaises(ApexTestException):
            task._check_code_coverage()

    def test_is_retriable_failure(self):
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "retry_failures": [
                "UNABLE_TO_LOCK_ROW",
                "unable to obtain exclusive access to this record",
            ],
        }
        task = RunApexTests(self.project_config, task_config, self.org_config)
        task._init_options(task_config.config["options"])

        self.assertTrue(
            task._is_retriable_failure(
                {
                    "Message": "UNABLE_TO_LOCK_ROW",
                    "StackTrace": "test",
                    "Outcome": "Fail",
                }
            )
        )
        self.assertTrue(
            task._is_retriable_failure(
                {
                    "Message": "TEST",
                    "StackTrace": "unable to obtain exclusive access to this record",
                    "Outcome": "Fail",
                }
            )
        )
        self.assertFalse(
            task._is_retriable_failure(
                {
                    "Message": "DUPLICATES_DETECTED",
                    "StackTrace": "test",
                    "Outcome": "Fail",
                }
            )
        )

    def test_init_options__regexes(self):
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "retry_failures": ["UNABLE_TO_LOCK_ROW"],
        }
        task = RunApexTests(self.project_config, task_config, self.org_config)
        task._init_options(task_config.config["options"])

        self.assertIsNotNone(
            task.options["retry_failures"][0].search("UNABLE_TO_LOCK_ROW: test failed")
        )

    def test_init_options__bad_regexes(self):
        task_config = TaskConfig()
        task_config.config["options"] = {
            "junit_output": "results_junit.xml",
            "poll_interval": 1,
            "test_name_match": "%_TEST",
            "retry_failures": ["("],
        }
        with self.assertRaises(TaskOptionsError):
            task = RunApexTests(self.project_config, task_config, self.org_config)
            task._init_options(task_config.config["options"])

    def test_get_namespace_filter__managed(self):
        task_config = TaskConfig({"options": {"managed": True, "namespace": "testns"}})
        task = RunApexTests(self.project_config, task_config, self.org_config)
        namespace = task._get_namespace_filter()
        self.assertEqual("'testns'", namespace)

    def test_get_namespace_filter__managed_no_namespace(self):
        task_config = TaskConfig({"options": {"managed": True}})
        task = RunApexTests(self.project_config, task_config, self.org_config)
        with self.assertRaises(TaskOptionsError):
            task._get_namespace_filter()

    def test_get_test_class_query__exclude(self):
        task_config = TaskConfig(
            {"options": {"test_name_match": "%_TEST", "test_name_exclude": "EXCL"}}
        )
        task = RunApexTests(self.project_config, task_config, self.org_config)
        query = task._get_test_class_query()
        self.assertEqual(
            "SELECT Id, Name FROM ApexClass WHERE NamespacePrefix = null "
            "AND (Name LIKE '%_TEST') AND (NOT Name LIKE 'EXCL')",
            query,
        )

    def test_run_task__no_tests(self):
        task = RunApexTests(self.project_config, self.task_config, self.org_config)
        task._get_test_classes = MagicMock(return_value={"totalSize": 0})
        task()
        self.assertIsNone(task.result)


@patch(
    "cumulusci.tasks.salesforce.BaseSalesforceTask._update_credentials",
    MagicMock(return_value=None),
)
class TestAnonymousApexTask(unittest.TestCase):
    def setUp(self):
        self.api_version = 42.0
        self.universal_config = UniversalConfig(
            {"project": {"api_version": self.api_version}}
        )
        self.tmpdir = tempfile.mkdtemp(dir=".")
        apex_path = os.path.join(self.tmpdir, "test.apex")
        with open(apex_path, "w") as f:
            f.write('System.debug("from file")')
        self.task_config = TaskConfig()
        self.task_config.config["options"] = {
            "path": apex_path,
            "apex": 'system.debug("Hello World!")',
            "param1": "StringValue",
        }
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )
        self.project_config.config = {
            "project": {
                "package": {"namespace": "abc", "api_version": self.api_version}
            }
        }
        keychain = BaseProjectKeychain(self.project_config, "")
        self.project_config.set_keychain(keychain)
        self.org_config = OrgConfig(
            {
                "id": "foo/1",
                "instance_url": "https://example.com",
                "access_token": "abc123",
                "namespace": "abc",
            },
            "test",
        )
        self.org_config._installed_packages = {}
        self.base_tooling_url = "{}/services/data/v{}/tooling/".format(
            self.org_config.instance_url, self.api_version
        )

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def _get_url_and_task(self):
        task = AnonymousApexTask(self.project_config, self.task_config, self.org_config)
        url = self.base_tooling_url + "executeAnonymous"
        return task, url

    def test_validate_options(self):
        task_config = TaskConfig({})
        with self.assertRaises(TaskOptionsError):
            AnonymousApexTask(self.project_config, task_config, self.org_config)

    def test_run_from_path_outside_repo(self):
        task_config = TaskConfig({"options": {"path": "/"}})
        task = AnonymousApexTask(self.project_config, task_config, self.org_config)
        with self.assertRaises(TaskOptionsError):
            task()

    def test_run_path_not_found(self):
        task_config = TaskConfig({"options": {"path": "bogus"}})
        task = AnonymousApexTask(self.project_config, task_config, self.org_config)
        with self.assertRaises(TaskOptionsError):
            task()

    def test_prepare_apex(self):
        self.task_config.config["options"]["namespaced"] = True

        task = AnonymousApexTask(self.project_config, self.task_config, self.org_config)
        before = "String %%%NAMESPACED_ORG%%%str = '%%%NAMESPACED_RT%%%';"
        expected = "String abc__str = 'abc.';"
        self.assertEqual(expected, task._prepare_apex(before))

    def test_prepare_apex__detect_namespace(self):
        task = AnonymousApexTask(self.project_config, self.task_config, self.org_config)
        before = "String %%%NAMESPACED_ORG%%%str = '%%%NAMESPACED_RT%%%';"
        expected = "String abc__str = 'abc.';"
        self.assertEqual(expected, task._prepare_apex(before))

    def test_optional_parameter_1_replacement(self):
        task = AnonymousApexTask(self.project_config, self.task_config, self.org_config)
        before = "String str = '%%%PARAM_1%%%';"
        expected = "String str = 'StringValue';"
        self.assertEqual(expected, task._prepare_apex(before))

    def test_optional_parameter_2_replacement(self):
        task = AnonymousApexTask(self.project_config, self.task_config, self.org_config)
        before = "String str = '%%%PARAM_2%%%';"
        expected = "String str = '';"
        self.assertEqual(expected, task._prepare_apex(before))

    @responses.activate
    def test_run_anonymous_apex_success(self):
        task, url = self._get_url_and_task()
        resp = {"compiled": True, "success": True}
        responses.add(responses.GET, url, status=200, json=resp)
        task()

    @responses.activate
    def test_run_string_only(self):
        task_config = TaskConfig({"options": {"apex": 'System.debug("test");'}})
        task = AnonymousApexTask(self.project_config, task_config, self.org_config)
        url = self.base_tooling_url + "executeAnonymous"
        responses.add(
            responses.GET, url, status=200, json={"compiled": True, "success": True}
        )
        task()

    @responses.activate
    def test_run_anonymous_apex_status_fail(self):
        task, url = self._get_url_and_task()
        responses.add(responses.GET, url, status=418, body="I'm a teapot")
        with self.assertRaises(SalesforceGeneralError) as cm:
            task()
        err = cm.exception
        self.assertEqual(str(err), "Error Code 418. Response content: I'm a teapot")
        self.assertTrue(err.url.startswith(url))
        self.assertEqual(err.status, 418)
        self.assertEqual(err.content, "I'm a teapot")

    @responses.activate
    def test_run_anonymous_apex_compile_except(self):
        task, url = self._get_url_and_task()
        problem = "Unexpected token '('."
        resp = {
            "compiled": False,
            "compileProblem": problem,
            "success": False,
            "line": 1,
            "column": 13,
            "exceptionMessage": "",
            "exceptionStackTrace": "",
            "logs": "",
        }
        responses.add(responses.GET, url, status=200, json=resp)
        with self.assertRaises(ApexCompilationException) as cm:
            task()
        err = cm.exception
        self.assertEqual(err.args[0], 1)
        self.assertEqual(err.args[1], problem)

    @responses.activate
    def test_run_anonymous_apex_except(self):
        task, url = self._get_url_and_task()
        problem = "Unexpected token '('."
        trace = "Line 0, Column 99"
        resp = {
            "compiled": True,
            "compileProblem": "",
            "success": False,
            "line": 1,
            "column": 13,
            "exceptionMessage": problem,
            "exceptionStackTrace": trace,
            "logs": "",
        }
        responses.add(responses.GET, url, status=200, json=resp)
        with self.assertRaises(ApexException) as cm:
            task()
        err = cm.exception
        self.assertEqual(err.args[0], problem)
        self.assertEqual(err.args[1], trace)

    @responses.activate
    def test_run_anonymous_apex__gack(self):
        task, url = self._get_url_and_task()
        responses.add(responses.GET, url, status=200, body="null")
        with self.assertRaises(SalesforceException) as cm:
            task()
        err = str(cm.exception)
        assert "gack" in err


@patch(
    "cumulusci.tasks.salesforce.BaseSalesforceTask._update_credentials",
    MagicMock(return_value=None),
)
class TestRunBatchApex(MockLoggerMixin, unittest.TestCase):
    def setUp(self):
        self.api_version = 42.0
        self.universal_config = UniversalConfig(
            {"project": {"api_version": self.api_version}}
        )
        self.task_config = TaskConfig()
        self.task_config.config["options"] = {
            "class_name": "ADDR_Seasonal_BATCH",
            "poll_interval": 1,
        }
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )
        self.project_config.config["project"] = {
            "package": {"api_version": self.api_version}
        }
        keychain = BaseProjectKeychain(self.project_config, "")
        self.project_config.set_keychain(keychain)
        self.org_config = OrgConfig(
            {
                "id": "foo/1",
                "instance_url": "https://example.com",
                "access_token": "abc123",
            },
            "test",
        )
        self.base_tooling_url = "{}/services/data/v{}/tooling/".format(
            self.org_config.instance_url, self.api_version
        )
        self.task_log = self._task_log_handler.messages

    def _get_query_resp(self):
        return {
            "size": 1,
            "totalSize": 1,
            "done": True,
            "queryLocator": None,
            "entityTypeName": "AsyncApexJob",
            "records": [
                {
                    "attributes": {
                        "type": "AsyncApexJob",
                        "url": "/services/data/v43.0/tooling/sobjects/AsyncApexJob/707L0000014nnPHIAY",
                    },
                    "Id": "707L0000014nnPHIAY",
                    "ApexClass": {
                        "attributes": {
                            "type": "ApexClass",
                            "url": "/services/data/v43.0/tooling/sobjects/ApexClass/01pL000000109ndIAA",
                        },
                        "Name": "ADDR_Seasonal_BATCH",
                    },
                    "Status": "Completed",
                    "ExtendedStatus": None,
                    "TotalJobItems": 1,
                    "JobItemsProcessed": 1,
                    "NumberOfErrors": 0,
                    "CreatedDate": "2018-08-07T16:00:56.000+0000",
                    "CompletedDate": "2018-08-07T16:01:57.000+0000",
                }
            ],
        }

    def _update_job_result(self, response: dict, result_dict: dict):
        "Extend the result from _get_query_resp with additional batch records"
        template_result = response["records"][-1]  # use the last result as a template
        assert isinstance(template_result, dict)
        new_result = {**template_result, **result_dict}  # copy with variations
        old_subjob_results = [  # set completed for all old subjob results
            {**record, "Status": "Completed"} for record in response["records"]
        ]
        result_list = [
            new_result
        ] + old_subjob_results  # prepend new result because SOQL is order by DESC
        return {**response, "records": result_list}

    def _get_url_and_task(self):
        task = BatchApexWait(self.project_config, self.task_config, self.org_config)
        url = (
            self.base_tooling_url
            + "query/?q=SELECT+Id%2C+ApexClass.Name%2C+Status%2C+ExtendedStatus%2C+TotalJobItems%2C+JobItemsProcessed%2C+NumberOfErrors%2C+CreatedDate%2C+CompletedDate+FROM+AsyncApexJob+WHERE+JobType+IN+%28%27BatchApex%27%2C%27Queueable%27%29+AND+ApexClass.Name%3D%27ADDR_Seasonal_BATCH%27++++ORDER+BY+CreatedDate+DESC++LIMIT+1+"
        )
        return task, url

    @responses.activate
    def test_run_batch_apex_status_fail(self):
        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"][0]["NumberOfErrors"] = 1
        response["records"][0]["ExtendedStatus"] = "Bad Status"
        responses.add(responses.GET, url, json=response)
        with self.assertRaises(SalesforceException) as cm:
            task()
        err = cm.exception
        self.assertIn("Bad Status", err.args[0])

    @responses.activate
    def test_run_batch_apex_number_mismatch(self):
        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"][0]["JobItemsProcessed"] = 1
        response["records"][0]["TotalJobItems"] = 3
        responses.add(responses.GET, url, json=response)
        task()

        self.assertIn("The final record counts do not add up.", self.task_log["info"])

    @responses.activate
    def test_run_batch_apex_status_ok(self):
        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        responses.add(responses.GET, url, json=response)
        task()

    @responses.activate
    def test_run_batch_apex_calc_elapsed_time(self):
        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        responses.add(responses.GET, url, json=response)
        task()
        self.assertEqual(task.elapsed_time(task.subjobs), 61)

    @responses.activate
    def test_run_batch_apex_queueable_status_failed(self):
        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"][0]["JobType"] = "Queueable"
        response["records"][0]["Status"] = "Failed"
        response["records"][0]["JobItemsProcessed"] = 0
        response["records"][0]["TotalJobItems"] = 0
        response["records"][0]["ExtendedStatus"] = "Error Details"
        responses.add(responses.GET, url, json=response)
        with self.assertRaises(SalesforceException) as e:
            task()
        assert "failure" in str(e.exception)

    @responses.activate
    def test_run_batch_apex_status_aborted(self):
        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"][0]["Status"] = "Aborted"
        response["records"][0]["JobItemsProcessed"] = 1
        response["records"][0]["TotalJobItems"] = 3
        responses.add(responses.GET, url, json=response)
        with self.assertRaises(SalesforceException) as e:
            task()
        assert "aborted" in str(e.exception)

    @responses.activate
    def test_run_batch_apex_status_failed(self):
        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"][0]["Status"] = "Failed"
        response["records"][0]["JobItemsProcessed"] = 1
        response["records"][0]["TotalJobItems"] = 3
        responses.add(responses.GET, url, json=response)
        self.task_log["info"] = []
        with self.assertRaises(SalesforceException) as e:
            task()
        assert "failure" in str(e.exception)

    @responses.activate
    def test_chained_subjobs(self):
        "Test subjobs that kick off a successor before they complete"
        task, url = self._get_url_and_task()
        url2 = (
            url.split("?")[0]
            + "?q=SELECT+Id%2C+ApexClass.Name%2C+Status%2C+ExtendedStatus%2C+TotalJobItems%2C+JobItemsProcessed%2C+NumberOfErrors%2C+CreatedDate%2C+CompletedDate+FROM+AsyncApexJob+WHERE+JobType+IN+%28%27BatchApex%27%2C%27Queueable%27%29+AND+ApexClass.Name%3D%27ADDR_Seasonal_BATCH%27++AND+CreatedDate+%3E%3D+2018-08-07T16%3A00%3A00Z++ORDER+BY+CreatedDate+DESC++"
        )

        # batch 1
        response = self._get_query_resp()
        batch_record = response["records"][0]
        batch_record.update(
            {
                "JobItemsProcessed": 1,
                "TotalJobItems": 3,
                "NumberOfErrors": 0,
                "Status": "Processing",
                "CreatedDate": "2018-08-07T16:00:00.000+0000",
            }
        )

        responses.add(responses.GET, url, json=response)

        # batch 2: 1 error
        response = self._update_job_result(
            response, {"Id": "Id2", "NumberOfErrors": 1, "Status": "Processing"}
        )
        responses.add(responses.GET, url2, json=response)

        # batch 3: found another error
        response = self._update_job_result(
            response, {"Id": "Id2", "NumberOfErrors": 2, "Status": "Processing"}
        )
        responses.add(responses.GET, url2, json=response)

        # batch 4: Complete, no errors in this sub-batch
        response = self._update_job_result(
            response,
            {
                "NumberOfErrors": 0,
                "Id": "Id4",
                "Status": "Completed",
                "CompletedDate": "2018-08-07T16:10:00.000+0000",  # 10 minutes passed
            },
        )
        responses.add(responses.GET, url2, json=response.copy())

        with self.assertRaises(SalesforceException) as e:
            task()

        assert len(task.subjobs) == 4
        summary = task.summarize_subjobs(task.subjobs)
        assert not summary["Success"]
        assert not summary["CountsAddUp"]
        assert summary["ElapsedTime"] == 10 * 60
        assert summary["JobItemsProcessed"] == 4
        assert summary["TotalJobItems"] == 12
        assert summary["NumberOfErrors"] == 3
        assert "batch errors" in str(e.exception)

    @responses.activate
    def test_chained_subjobs_beginning(self):
        "Test the first subjob that kicks off a successor before they complete"
        task, url = self._get_url_and_task()
        url2 = (
            url.split("?")[0]
            + "?q=SELECT+Id%2C+ApexClass.Name%2C+Status%2C+ExtendedStatus%2C+TotalJobItems%2C+JobItemsProcessed%2C+NumberOfErrors%2C+CreatedDate%2C+CompletedDate+FROM+AsyncApexJob+WHERE+JobType+IN+%28%27BatchApex%27%2C%27Queueable%27%29+AND+ApexClass.Name%3D%27ADDR_Seasonal_BATCH%27++AND+CreatedDate+%3E%3D+2018-08-07T16%3A00%3A00Z++ORDER+BY+CreatedDate+DESC++"
        )

        # batch 1
        response = self._get_query_resp()
        responses.add(responses.GET, url2, json=response)

        batch_record = response["records"][0]
        batch_record.update(
            {
                "JobItemsProcessed": 1,
                "TotalJobItems": 3,
                "NumberOfErrors": 0,
                "Status": "Preparing",
                "CreatedDate": "2018-08-07T16:00:00.000+0000",
                "CompletedDate": None,
            }
        )

        real_poll_action = task._poll_action
        counter = 0

        def mock_poll_action():
            nonlocal counter
            counter += 1
            if counter == 1:
                task.poll_complete = False
                return real_poll_action()
            else:
                rc = real_poll_action()
                task.poll_complete = True
                return rc

        task._poll_action = mock_poll_action

        responses.add(responses.GET, url, json=response)

        task()
        assert counter == 2
        assert task.poll_complete
        summary = task.summarize_subjobs(task.subjobs)
        assert not summary["NumberOfErrors"]

    @responses.activate
    def test_chained_subjobs_halfway(self):
        "Test part-way through a series of subjobs that kick off a successor before they complete"
        task, url = self._get_url_and_task()
        url2 = (
            url.split("?")[0]
            + "?q=SELECT+Id%2C+ApexClass.Name%2C+Status%2C+ExtendedStatus%2C+TotalJobItems%2C+JobItemsProcessed%2C+NumberOfErrors%2C+CreatedDate%2C+CompletedDate+FROM+AsyncApexJob+WHERE+JobType+IN+%28%27BatchApex%27%2C%27Queueable%27%29+AND+ApexClass.Name%3D%27ADDR_Seasonal_BATCH%27++AND+CreatedDate+%3E%3D+2018-08-07T16%3A00%3A00Z++ORDER+BY+CreatedDate+DESC++"
        )

        # batch 1
        response = self._get_query_resp()
        responses.add(responses.GET, url2, json=response)

        batch_record = response["records"][0]
        batch_record.update(
            {
                "JobItemsProcessed": 1,
                "TotalJobItems": 3,
                "NumberOfErrors": 0,
                "Status": "Preparing",
                "CreatedDate": "2018-08-07T16:00:00.000+0000",
                "CompletedDate": "2018-08-07T16:05:00.000+0000",
            }
        )

        # batch 2: 1 error
        response = self._update_job_result(
            response, {"Id": "Id2", "NumberOfErrors": 1, "CompletedDate": None}
        )
        responses.add(responses.GET, url2, json=response)

        real_poll_action = task._poll_action
        counter = 0

        def mock_poll_action():
            nonlocal counter
            counter += 1
            if counter == 1:
                task.poll_complete = False
                return real_poll_action()
            else:
                rc = real_poll_action()
                task.poll_complete = True
                return rc

        task._poll_action = mock_poll_action

        responses.add(responses.GET, url, json=response)

        task()
        assert counter == 2
        assert task.poll_complete
        summary = task.summarize_subjobs(task.subjobs)
        assert not summary["NumberOfErrors"]

    @responses.activate
    def test_job_not_found(self):

        task, url = self._get_url_and_task()
        response = self._get_query_resp()
        response["records"] = []
        responses.add(responses.GET, url, json=response)

        with self.assertRaises(SalesforceException) as e:
            task()

        assert "found" in str(e.exception)


class TestApexIntegrationTests:
    @pytest.mark.org_shape("qa", "ccitest:qa_org")
    @pytest.mark.slow()
    @pytest.mark.skip()  # until our CI has access to github service, or test
    # doesn't rely on it. In the meantime, the VCR test below is still good
    # and this test works on laptops.
    def test_run_tests__integration_test__call_salesforce(self, create_task, caplog):
        self._test_run_tests__integration_test(create_task, caplog)

    # There were challenges VCR-ing this because it depends on a
    # particular flow. To recreate the tape you'll need to run the
    # ccitest:qa_org flow aginst your org first.
    #
    # Also: some redundant "polls" were removed by hand to avoid
    # disk space usage.
    def test_run_tests__integration_test(self, create_task, caplog, vcr):
        with vcr.use_cassette(
            "ManualEditTestApexIntegrationTests.test_run_tests__integration_test.yaml",
            record_mode="none",
        ):
            self._test_run_tests__integration_test(create_task, caplog)

    def _test_run_tests__integration_test(self, create_task, caplog):

        caplog.set_level(logging.INFO)
        with pytest.raises(exc.ApexTestException) as e:
            task = create_task(
                RunApexTests,
                {
                    "required_org_code_coverage_percent": 70,
                    "required_per_class_code_coverage_percent": 60,
                    "json_output": None,
                    "junit_output": None,
                },
            )
            with patch.object(task, "_update_credentials"):
                task()
        relevant_records = [
            record for record in caplog.records if "below required level" in str(record)
        ]
        assert len(relevant_records) == 1, caplog.records
        assert "SampleClass2" in str(relevant_records[0])
        assert "below required level" in str(e.value)
        assert "SampleClass2" in str(e.value)

        caplog.clear()

        task = create_task(
            RunApexTests,
            {
                "required_org_code_coverage_percent": 70,
                "required_per_class_code_coverage_percent": 40,
                "json_output": None,
                "junit_output": None,
            },
        )
        with patch.object(task, "_update_credentials"):
            task()
        relevant_records = [
            record for record in caplog.records if "expectations" in str(record)
        ]
        assert len(relevant_records) == 2
        assert "All classes meet" in str(relevant_records)
        assert "Organization-wide code" in str(relevant_records)



import csv
import time
from datetime import datetime, timedelta

from dateutil import tz
from dateutil.parser import isoparse

from cumulusci.core.exceptions import (
    CumulusCIException,
    PushApiObjectNotFound,
    TaskOptionsError,
)
from cumulusci.core.utils import process_bool_arg
from cumulusci.tasks.push.push_api import SalesforcePushApi
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class BaseSalesforcePushTask(BaseSalesforceApiTask):
    completed_statuses = ["Succeeded", "Failed", "Canceled"]
    api_version = "38.0"

    def _init_task(self):
        super(BaseSalesforcePushTask, self)._init_task()
        self.push = SalesforcePushApi(self.sf, self.logger)

    def _parse_version(self, version):
        # Parse the version number string
        major = None
        minor = None
        patch = None
        build = None
        state = "Released"
        version_parts = version.split(".")
        if len(version_parts) >= 1:
            major = version_parts[0]
        if len(version_parts) == 2:
            minor = version_parts[1]
            if minor.find("Beta") != -1:
                state = "Beta"
                minor, build = minor.replace(" (Beta ", ",").replace(")", "").split(",")
        if len(version_parts) > 2:
            minor = version_parts[1]
            patch = version_parts[2]
            if patch.find("Beta") != -1:
                state = "Beta"
                patch, build = minor.replace(" (Beta ", ",").replace(")", "").split(",")

        return {
            "major": major,
            "minor": minor,
            "patch": patch,
            "build": build,
            "state": state,
        }

    def _get_version(self, package, version):

        version_info = self._parse_version(version)

        version_where = (
            "ReleaseState = '{}'".format(version_info["state"])
            + " AND MajorVersion = {}".format(version_info["major"])
            + " AND MinorVersion = {}".format(version_info["minor"])
        )
        if version_info.get("patch"):
            version_where += " AND PatchVersion = {}".format(version_info["patch"])
        if version_info["state"] == "Beta" and version_info.get("build"):
            version_where += " AND BuildNumber = {}".format(version_info["build"])

        version = package.get_package_version_objs(version_where, limit=1)
        if not version:
            raise PushApiObjectNotFound(
                "PackageVersion not found. "
                + f"Namespace = {package.namespace}, Version Info = {version_info}"
            )
        return version[0]

    def _get_package(self, metadata_package_id: str = None, namespace: str = None):
        if metadata_package_id:
            where_clause = f"Id = '{metadata_package_id}'"
        else:
            where_clause = f"NamespacePrefix = '{namespace}'"

        package = self.push.get_package_objs(where_clause, limit=1)

        if not package:
            raise PushApiObjectNotFound(
                f"No package was found using the following query: {where_clause}"
            )

        return package[0]

    def _load_orgs_file(self, path):
        orgs = []
        with open(path, "r") as f:
            for line in f:
                if line.isspace():
                    continue
                orgs.append(line.split()[0])
        return orgs

    def _get_push_request_query(self, request_id):
        default_where = {"PackagePushRequest": "Id = '{}'".format(request_id)}

        # Create a new PushAPI instance with different settings than self.push
        self.push_report = SalesforcePushApi(
            self.sf,
            self.logger,
            lazy=["subscribers", "jobs"],
            default_where=default_where,
        )

        # Get the push request
        self.push_request = self.push_report.get_push_request_objs(
            "Id = '{}'".format(request_id), limit=1
        )
        if not self.push_request:
            raise PushApiObjectNotFound(
                "Push Request {} was not found".format(self.push_request)
            )

        self.push_request = self.push_request[0]

    def _get_push_request_job_results(self):
        failed_jobs = []
        success_jobs = []
        canceled_jobs = []

        jobs = self.push_request.get_push_job_objs()
        for job in jobs:
            if job.status == "Failed":
                failed_jobs.append(job)
            elif job.status == "Succeeded":
                success_jobs.append(job)
            elif job.status == "Canceled":
                canceled_jobs.append(job)

        self.logger.info(
            "Push complete: {} succeeded, {} failed, {} canceled".format(
                len(success_jobs), len(failed_jobs), len(canceled_jobs)
            )
        )

        failed_by_error = {}
        for job in failed_jobs:
            errors = job.get_push_error_objs()
            for error in errors:
                error_key = (
                    error.error_type,
                    error.title,
                    error.message,
                    error.details,
                )
                if error_key not in failed_by_error:
                    failed_by_error[error_key] = []
                failed_by_error[error_key].append(error)

        if failed_jobs:
            self.logger.info("-----------------------------------")
            self.logger.info("Failures by error type")
            self.logger.info("-----------------------------------")
            for key, errors in failed_by_error.items():
                self.logger.info("    ")
                self.logger.info("{} failed with...".format(len(errors)))
                self.logger.info("    Error Type = {}".format(key[0]))
                self.logger.info("    Title = {}".format(key[1]))
                self.logger.info("    Message = {}".format(key[2]))
                self.logger.info("    Details = {}".format(key[3]))

    def _report_push_status(self, request_id):
        self._get_push_request_query(request_id)
        # Check if the request is complete
        interval = 10
        if self.push_request.status not in self.completed_statuses:
            self.logger.info(
                "Push request is not yet complete."
                + " Polling for status every {} seconds until completion".format(
                    interval
                )
            )

        # Loop waiting for request completion
        i = 0
        while self.push_request.status not in self.completed_statuses:
            if i == 10:
                self.logger.info("This is taking a while! Polling every 60 seconds")
                interval = 60
            time.sleep(interval)

            # Clear the method level cache on get_push_requests and
            # get_push_request_objs
            self.push_report.get_push_requests.cache_clear()
            self.push_report.get_push_request_objs.cache_clear()
            # Get the push_request again
            self.push_request = self.push_report.get_push_request_objs(
                "Id = '{}'".format(request_id), limit=1
            )[0]
            self.logger.info(self.push_request.status)
            i += 1

        self._get_push_request_job_results()


class SchedulePushOrgList(BaseSalesforcePushTask):

    task_options = {
        "csv": {"description": "The path to a CSV file to read.", "required": False},
        "csv_field_name": {
            "description": "The CSV field name that contains organization IDs. Defaults to 'OrganizationID'",
            "required": False,
        },
        "orgs": {
            "description": "The path to a file containing one OrgID per line.",
            "required": False,
        },
        "version": {
            "description": "The managed package version to push",
        },
        "version_id": {
            "description": "The MetadataPackageVersionId (ID prefix `04t`) to push",
        },
        "metadata_package_id": {
            "description": "The MetadataPackageId (ID prefix `033`) to push.",
        },
        "namespace": {
            "description": (
                "The managed package namespace to push."
                + " Defaults to project__package__namespace."
            )
        },
        "start_time": {
            "description": (
                "Set the start time (ISO-8601) to queue a future push."
                " (Ex: 2021-01-01T06:00Z or 2021-01-01T06:00-08:00)"
                " Times with no timezone will be interpreted as UTC."
            )
        },
        "batch_size": {
            "description": (
                "Break pull requests into batches of this many orgs."
                + " Defaults to 200."
            )
        },
    }

    def _init_task(self):
        super(SchedulePushOrgList, self)._init_task()
        self.push = SalesforcePushApi(self.sf, self.logger, self.options["batch_size"])

    def _init_options(self, kwargs):
        super(SchedulePushOrgList, self)._init_options(kwargs)
        self.options["dry_run"] = False
        neither_file_option = "orgs" not in self.options and "csv" not in self.options
        both_file_options = "orgs" in self.options and "csv" in self.options
        if neither_file_option or both_file_options:
            raise TaskOptionsError(
                "Please call this task with either the `orgs` or `csv` option."
            )
        # Set the namespace option to the value from cumulusci.yml if not
        # already set
        if "namespace" not in self.options:
            self.options["namespace"] = self.project_config.project__package__namespace
        if "metadata_package_id" not in self.options:
            self.options[
                "metadata_package_id"
            ] = self.project_config.project__package__metadata_package_id
        if "batch_size" not in self.options:
            self.options["batch_size"] = 200
        if "csv" not in self.options and "csv_field_name" in self.options:
            raise TaskOptionsError("Please provide a csv file for this task to run.")

    def _get_orgs(self):
        if "csv" in self.options:
            with open(self.options.get("csv"), newline="", encoding="utf-8") as csvfile:
                reader = csv.DictReader(csvfile)
                return [
                    row[self.options.get("csv_field_name", "OrganizationId")]
                    for row in reader
                ]
        else:
            return self._load_orgs_file(self.options.get("orgs"))

    def _run_task(self):
        orgs = self._get_orgs()

        package_id = self.options.get("metadata_package_id")
        version_id = self.options.get("version_id")
        namespace_opt = self.options.get("namespace")

        package = self._get_package(
            metadata_package_id=package_id,
            namespace=namespace_opt,
        )
        if version_id:
            version = version_id
        else:
            version = self._get_version(package, self.options.get("version")).sf_id

        utcnow = datetime.now(tz.UTC)
        start_time = self.options.get("start_time")
        if start_time:
            if start_time.lower() == "now":
                start_time = utcnow + timedelta(seconds=5)
            else:
                start_time = isoparse(start_time)
                if start_time.utcoffset() is None:
                    start_time = start_time.replace(tzinfo=tz.UTC)
            if start_time < utcnow:
                raise CumulusCIException("Start time cannot be in the past")
        else:
            # delay a bit to allow for review
            delay_minutes = 5
            self.logger.warning(
                "Scheduling push for %d minutes from now", delay_minutes
            )
            start_time = utcnow + timedelta(minutes=delay_minutes)

        if self.options["dry_run"]:
            self.logger.info(
                f"Selected {len(orgs)} orgs. "
                f"Skipping actual creation of the PackagePushRequest because the dry_run flag is on."
            )
            return

        self.request_id, num_scheduled_orgs = self.push.create_push_request(
            version, orgs, start_time
        )

        self.return_values["request_id"] = self.request_id

        if num_scheduled_orgs > 1000:
            sleep_time_s = 30
            self.logger.info(
                "Delaying {} seconds to allow all jobs to initialize".format(
                    sleep_time_s
                )
            )
            time.sleep(sleep_time_s)
        elif num_scheduled_orgs == 0:
            self.logger.warning("Canceling push request with 0 orgs")
            self.push.cancel_push_request
            return

        self.logger.info("Setting status to Pending to queue execution.")
        self.logger.info(f"The push upgrade will start at {start_time}")

        # Run the job
        self.logger.info(self.push.run_push_request(self.request_id))
        self.logger.info(
            "Push Request {} is queued for execution.".format(self.request_id)
        )

        # Report the status if start time is less than 1 minute from now
        if start_time - utcnow < timedelta(minutes=1):
            self._report_push_status(self.request_id)
        else:
            self.logger.info("Exiting early since request is in the future")


class SchedulePushOrgQuery(SchedulePushOrgList):
    task_options = {
        "version": {
            "description": "The managed package version to push",
            "required": True,
        },
        "subscriber_where": {
            "description": (
                "A SOQL style WHERE clause for filtering PackageSubscriber"
                + " objects. Ex: OrgType = 'Sandbox'"
            )
        },
        "min_version": {
            "description": (
                "If set, no subscriber with a version lower than min_version"
                + " will be selected for push"
            )
        },
        "metadata_package_id": {
            "description": "The MetadataPackageId (ID prefix `033`) to push.",
            "required": False,
        },
        "namespace": {
            "description": (
                "The managed package namespace to push."
                + " Defaults to project__package__namespace."
            )
        },
        "start_time": {
            "description": (
                "Set the start time (ISO-8601) to queue a future push."
                " (Ex: 2021-01-01T06:00Z or 2021-01-01T06:00-08:00)"
                " Times with no timezone will be interpreted as UTC."
            )
        },
        "dry_run": {
            "description": "If True, log how many orgs were selected but skip creating a PackagePushRequest.  Defaults to False"
        },
    }

    def _init_options(self, kwargs):
        super(SchedulePushOrgList, self)._init_options(kwargs)
        # Set the namespace option to the value from cumulusci.yml if not
        # already set
        if "namespace" not in self.options:
            self.options["namespace"] = self.project_config.project__package__namespace
        if "batch_size" not in self.options:
            self.options["batch_size"] = 200
        self.options["dry_run"] = process_bool_arg(self.options.get("dry_run", False))

    def _get_orgs(self):
        subscriber_where = self.options.get("subscriber_where")
        default_where = {
            "PackageSubscriber": ("OrgStatus != 'Inactive' AND InstalledStatus = 'i'")
        }
        if subscriber_where:
            default_where["PackageSubscriber"] += " AND ({})".format(subscriber_where)

        push_api = SalesforcePushApi(
            self.sf, self.logger, default_where=default_where.copy(), bulk=self.bulk
        )

        package = self._get_package(
            metadata_package_id=self.options.get("metadata_package_id"),
            namespace=self.options.get("namespace"),
        )
        version = self._get_version(package, self.options.get("version"))
        min_version = self.options.get("min_version")
        if min_version:
            min_version = self._get_version(package, self.options.get("min_version"))

        orgs = []

        if min_version:
            # If working with a range of versions, use an inclusive search
            versions = version.get_older_released_version_objs(min_version=min_version)
            included_versions = []
            for include_version in versions:
                included_versions.append(str(include_version.sf_id))
            if not included_versions:
                raise ValueError(
                    "No versions found between version id {} and {}".format(
                        version.version_number, min_version.version_number
                    )
                )

            # Query orgs for each version in the range individually to avoid
            # query timeout errors with querying multiple versions
            for included_version in included_versions:
                # Clear the get_subscribers method cache before each call
                push_api.get_subscribers.cache_clear()
                push_api.default_where[
                    "PackageSubscriber"
                ] = "{} AND MetadataPackageVersionId = '{}'".format(
                    default_where["PackageSubscriber"], included_version
                )
                for subscriber in push_api.get_subscribers():
                    orgs.append(subscriber["OrgKey"])

        else:
            # If working with a specific version rather than a range, use an
            # exclusive search.
            # Add exclusion of all orgs running on newer releases
            newer_versions = version.get_newer_released_version_objs()
            excluded_versions = [str(version.sf_id)]
            for newer in newer_versions:
                excluded_versions.append(str(newer.sf_id))
            push_api.default_where[
                "PackageSubscriber"
            ] += " AND MetadataPackageVersionId NOT IN {}".format(
                "('" + "','".join(excluded_versions) + "')"
            )

            for subscriber in push_api.get_subscribers():
                orgs.append(subscriber["OrgKey"])

        return orgs



import os
import unittest
from unittest import mock

from cumulusci.core.config import BaseProjectConfig, TaskConfig, UniversalConfig
from cumulusci.core.dependencies.dependencies import PackageNamespaceVersionDependency
from cumulusci.tasks.metaxml import UpdateApi, UpdateDependencies
from cumulusci.utils import temporary_dir


class TestUpdateApi(unittest.TestCase):
    def test_run_task(self):
        with temporary_dir() as d:
            os.mkdir(".git")
            os.mkdir("src")
            meta_xml_path = os.path.join(d, "src", "test-meta.xml")
            with open(meta_xml_path, "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <apiVersion>1.0</apiVersion>
</ApexClass>
"""
                )

            project_config = BaseProjectConfig(
                UniversalConfig(), config={"noyaml": True}
            )
            task_config = TaskConfig({"options": {"version": "43.0"}})
            task = UpdateApi(project_config, task_config)
            task()

            with open(meta_xml_path, "r") as f:
                result = f.read()
            self.assertEqual(
                """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <apiVersion>43.0</apiVersion>
</ApexClass>
""",
                result,
            )


class TestUpdateDependencies(unittest.TestCase):
    @mock.patch("cumulusci.tasks.metaxml.get_static_dependencies")
    def test_run_task(self, get_static_dependencies):
        get_static_dependencies.return_value = [
            PackageNamespaceVersionDependency(namespace="npe01", version="1.1"),
            PackageNamespaceVersionDependency(namespace="npsp", version="3.0"),
        ]

        with temporary_dir() as d:
            os.mkdir(".git")
            os.mkdir("src")
            meta_xml_path = os.path.join(d, "src", "test-meta.xml")
            with open(meta_xml_path, "w") as f:
                f.write(
                    """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <packageVersions>
        <namespace>npsp</namespace>
        <majorNumber>1</majorNumber>
        <minorNumber>0</minorNumber>
    </packageVersions>
    <packageVersions>
        <namespace>npe01</namespace>
        <majorNumber>1</majorNumber>
        <minorNumber>0</minorNumber>
    </packageVersions>
</ApexClass>
"""
                )

            project_config = BaseProjectConfig(
                UniversalConfig(), config={"noyaml": True}
            )
            task_config = TaskConfig()
            task = UpdateDependencies(project_config, task_config)
            task()

            with open(meta_xml_path, "r") as f:
                result = f.read()
            self.assertEqual(
                """<?xml version="1.0" encoding="UTF-8"?>
<ApexClass xmlns="http://soap.sforce.com/2006/04/metadata">
    <packageVersions>
        <namespace>npsp</namespace>
        <majorNumber>3</majorNumber>
        <minorNumber>0</minorNumber>
    </packageVersions>
    <packageVersions>
        <namespace>npe01</namespace>
        <majorNumber>1</majorNumber>
        <minorNumber>1</minorNumber>
    </packageVersions>
</ApexClass>
""",
                result,
            )
